/*
 This file is part of DepQBF.

 DepQBF, a solver for quantified boolean formulae (QBF).        
 Copyright 2017, 2018 Viktor Schuppan.
 Copyright 2013, 2014, 2015, 2016, 2017 Florian Lonsing,
   Vienna University of Technology, Vienna, Austria.
 Copyright 2010, 2011, 2012 Florian Lonsing,
   Johannes Kepler University, Linz, Austria.
 Copyright 2012 Aina Niemetz,
   Johannes Kepler University, Linz, Austria.

 DepQBF is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or (at
 your option) any later version.

 DepQBF is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with DepQBF.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>
#include <dirent.h>
#include <assert.h>
#include <string.h>
#include <stdarg.h>
#include <signal.h>
#include <unistd.h>
#include <unistd.h>
#include "qdpll.h"
#include "qdpll_internals.h"

#define VERSION                                                         \
  "DepQBF 6.03\n"                                                        \
  "Copyright 2017 Viktor Schuppan.\n"                                   \
  "Copyright 2013, 2014, 2015, 2016, 2017 Florian Lonsing,\n"           \
  "  Vienna University of Technology, Vienna, Austria.\n"               \
  "Copyright 2010, 2011, 2012 Florian Lonsing,\n"                       \
  "  Johannes Kepler University, Linz, Austria.\n"                       \
  "Copyright 2012 Aina Niemetz,\n"                                       \
  "  Johannes Kepler University, Linz, Austria.\n"                        \
  "This is free software; see COPYING for copying conditions.\n"        \
  "There is NO WARRANTY, to the extent permitted by law.\n"

#define USAGE1 \
"usage: depqbf-a2aecc [ <option> ... ] [ <in-file> ]\n"\
"\n"\
"  where <in-file> is a file in (Q)DIMACS format (default: stdin)\n"\
"  and <option> is any combination of the following:\n"\
"\n"\
"  Note: see function 'qdpll_configure' in file 'qdpll.c' for further, undocumented\n"\
"    options which are not listed here. However, many of those are experimental.\n"\
"\n"\
"  -h, --help                      display usage information\n"\
"  -v                              enable verbosity incrementally (at most '-v -v')\n"\
"  --pretty-print                  print the parsed formula after cleaning up (delete tautologies,\n"\
"                                    superfluous variables, quantifier blocks, and literals; sort \n"\
"                                    literals in clauses by quantifier nesting levels and IDs).\n"\
"  --print-stats                   prints statistics about the parsed formula and, if enabled, about\n"\
"                                  an unsatisfiable core.\n"\
"  --incremental-use               Relevant ONLY for API use: enable incremental solving. \n"\
 "                                   Must be combined with '--dep-man=simple'. \n"      \
"  --qdo                           QDIMACS output generation (partial certificate):\n"\
"                                    If the outermost (i.e. leftmost) quantifier block\n"\
"                                    of a satisfiable QBF is existentially quantified,\n"\
"                                    then print an assignment to the variables of this\n"\
"                                    block (and dual for unsatisfiable QBFs and\n"\
"                                    universal variables from the outermost block,\n"\
"                                    if that block is universally quantified).\n"\
"                                    IMPORTANT: must be combined with '--no-dynamic-nenofex'\n"\
"  --traditional-qcdcl             apply a traditional variant of clause and cube learning (QCDCL),\n"\
"                                    which was applied in previous versions of DepQBF.\n"\
"                                    In this version, by default lazy QPUP-based QCDCL is applied.\n"\
"  --long-dist-res                 Apply long-distance resolution in constraint learning. \n"\
"                                    Should be combined with '--dep-man=simple'.\n"\
"                                    Optionally, '--traditional-qcdcl' may be specified which enables\n"\
"                                    long-distance resolution where intermediate resolutions are carried\n"\
"                                    out starting from the empty clause.\n" \
"  --no-lazy-qpup                  disable lazy QPUP-based QCDCL and carry out all resolution\n"\
"                                    steps starting from a separating cut in the implication graph.\n"\
"                                    This works also with long-distance resolution.\n"                   \
"  --no-qpup-cdcl                  apply traditional QCDCL for clause learning (instead of QPUP).\n"\
"  --no-qpup-sdcl                  apply traditional QCDCL for cube learning (instead of QPUP).\n"\
"  --trace[=<format>]              dump trace in <format> to <stdout>\n"\
"                                    format: qrp  ... ascii QRP format (default)\n"\
"                                            bqrp ... binary QRP format\n"\
"                                    NOTE: tracing must be combined with options '--dep-man=simple' and\n"\
"                                          either '--traditional-qcdcl' (which disables QPUP-based QCDCL)\n"\
"                                          or '--no-lazy-qpup' (to enable tracing with QPUP-based QCDCL).\n"

#define USAGE2 \
"  --dep-man=<val>                 set dependency manager: if <val>=qdag (default) then the solver\n"\
"                                    uses the standard dependency scheme; if <val>=simple then the\n"\
"                                    solver uses the given quantifier prefix of the input formula\n"	\
"  --no-cdcl                       disable conflict-driven clause learning and\n"\
"                                    backtrack chronologically from conflicts\n"\
"  --no-sdcl                       disable solution-driven cube learning and\n"\
"                                    backtrack chronologically from solutions\n"\
"  --no-pure-literals              disable pure literal detection\n"\
"  --no-spure-literals             DEPRECATED: include ALL constraints for pure literal detection\n"\
"  --no-unit-mtf                   no move-to-front (MTF) of learnt constraints which became unit\n"\
"  --no-res-mtf                    no move-to-front (MTF) of learnt constraints which became empty\n"\
"  --lclauses-init-size=<val>      initially allow <val> clauses to be learned before resizing the clause list\n"\
"  --lcubes-init-size=<val>        initially allow <val> cubes to be learned before resizing the cube list\n"\
"  --lclauses-resize-value=<val>   increase capacity of learned clauses list by <val>\n"\
"  --lcubes-resize-value=<val>     increase capacity of learned cubes list by <val>\n"\
"  --orestart-dist-init=<val>      set initial distance of outer restarts to <val> (default 10)\n"\
"  --orestart-dist-inc=<val>       increase distance of outer restarts by <val> (default 5)\n"\
"  --irestart-dist-init=<val>      set initial distance of inner restarts to <val> (default 100)\n"\
"  --irestart-dist-inc=<val>       increase distance of inner restarts by <val> (default 10)\n"\
"  --max-dec=<val>                 Abort after <val> assignments by decision making.\n"\
"  --max-btracks=<val>             Abort after <val> backtracks.\n"\
"  --max-secs=<val>                Abort after <val> seconds.\n"\
"\n"\
"Options that control QBCE and applications of generalized axioms:"\
"\n"\
"  --no-qbce-dynamic               disable dynamic QBCE (enabled by default)\n"\
"  --qbce-preprocessing            enable QBCE preprocessing (must be preceded by '--no-qbce-dynamic')\n"\
"  --qbce-inprocessing             enable QBCE inprocessing (must be preceded by '--no-qbce-dynamic')\n"\
"  --qbce-witness-max-occs=<val>   maximum number <val> of occurrences considered in QBCE (default: 50)\n"\
"  --qbce-max-clause-size=<val>    maximum length <val> of clauses considered in QBCE (default: 50)\n"\
"  --no-empty-formula-watching     disable empty formula watching (default: enabled)\n"\
"  --no-dynamic-nenofex            disable dynamic nenofex tests (default: enabled)\n"\
"  --dyn-nenofex-ignore-unsat      ignore unsatisfiable branch detected by nenofex\n"\
"  --dyn-nenofex-ignore-sat        ignore satisfiable branch detected by nenofex\n"\
"  --no-trivial-falsity            disable trivial falsity tests (default: enabled)\n"\
"  --no-trivial-truth              disable trivial truth tests (default: enabled)\n"\
"\n"

#define USAGE3 \
"Options for unsatisfiable core (uc) extraction:"\
"\n"\
"  --uc={none,c,q,qc,              'none' disables uc extraction (default).\n"\
"        enuma2e1,enuma2e2}        'c', 'q', 'qc', 'enuma2e1', and 'enuma2e2' enable uc extraction\n"\
"                                  (requires '--incremental-use').\n"\
"                                  'c' removes clauses from the matrix,\n"\
"                                  'q' changes universal to existential quantifiers,\n"\
"                                  'qc' removes clauses from the matrix and additionally changes\n"\
"                                  universal to existential quantifiers.\n"\
"                                  'enuma2e1', 'enuma2e2' exhaustively enumerate universally quantified\n"\
"                                  variables that can be weakened individually to existentially quantified\n"\
"                                  variables while preserving unsatisfiability\n"\
"  --min-uc                        with '--uc={c,q,qc}' enable minimization of uc (default: disabled)\n"\
"  --uc-separate-csr               with '--uc=qc --min-uc' restrict clause set refinement during the\n"\
"                                  initial uc computation and during the minimization of quantifiers to\n"\
"                                  quantifiers instead of performing clause set refinement on both\n"\
"                                  quantifiers and matrix\n"\
"  --uc-file <uc-file>             write uc to file <uc-file> (default: disabled) (requires '--uc={c,q,qc}')\n"\
"  --print-a2aecc <a2aecc-file>    write a2aecc transformed file to <a2aecc-file> and exit\n"\
"  --a2aecc-prefix-as-sets         optimizes the a2aecc encoding by treating a sequence of universal\n"\
"                                  quantifiers as a set (may affect semantics) (default: disabled)\n"\
"                                  (requires '--uc={q,qc,enuma2e1,enuma2e2} or --print-a2aecc\n"\
"  --a2aecc-poldep-clauses         optimizes the a2aecc encoding by generating a2aecc clauses only as\n"\
"                                  required by polarity (default: disabled) (requires\n"\
"                                  '--uc={q,qc,enuma2e1,enuma2e2} or --print-a2aecc\n"\
"  --enuma2e-witness-base <base>   write enuma2e1/2 witnesses to files with basename <base>\n"\
"                                   (requires '--uc={enuma2e1,enuma2e2}')\n"\
"\n"


typedef struct {
  int size;
  ClauseGroupID *keys;
} cgidhash_t;

static cgidhash_t *cgidhash_new (QDPLLMemMan *mm, int size) {
  int nonnullsize = (size == 0) ? 1 : size;
  cgidhash_t *h = qdpll_malloc(mm, sizeof (cgidhash_t));
  memset (h, 0, sizeof (cgidhash_t));
  h->keys = qdpll_malloc (mm, nonnullsize * sizeof (ClauseGroupID));
  memset (h->keys, 0, nonnullsize * sizeof (ClauseGroupID));
  h->size = nonnullsize;
  return h;
}

int cgidhash_index (cgidhash_t *h, ClauseGroupID key) {
  int i = (int) key % h->size;
  while (h->keys[i] && h->keys[i] != key)
    i = (i + 1) % h->size;
  return i;
}

void cgidhash_insert (cgidhash_t *h, ClauseGroupID key) {
  int i = cgidhash_index(h, key);
  h->keys[i] = key;
}

int cgidhash_present (cgidhash_t *h, ClauseGroupID key) {
  int i = cgidhash_index(h, key);
  return (h->keys[i] != 0);
}

static void cgidhash_delete (QDPLLMemMan *mm, cgidhash_t *h) {
  qdpll_free(mm, h->keys, h->size * sizeof (ClauseGroupID));
  qdpll_free(mm, h, sizeof (cgidhash_t));
}

#define QDPLL_A2AECC_ABORT_APP(cond,msg) \
  do {                                  \
    if (cond)                                                                \
      {                                                                        \
        fprintf (stderr, "[QDPLL-A2AECC-APP] %s at line %d: %s\n", __func__,        \
                 __LINE__, msg);                                        \
        fflush (stderr);                                                \
        abort();                                                        \
      }                                                                        \
  } while (0)


#define QDPLL_A2AECC_INIT_LEN_COMMENTS 1000

#define QDPLL_A2AECC_INIT_LEN_VAROCCCLAUSEGROUPIDS 2

struct QDPLLA2AECCVarOccCClauseGroupIDs
{
  ClauseGroupID *varocc_clausegroupids,
    *varocc_clausegroupids_next,
    *varocc_clausegroupids_end;
};

typedef struct QDPLLA2AECCVarOccCClauseGroupIDs QDPLLA2AECCVarOccCClauseGroupIDs;

struct QDPLLA2AECCApp
{
  struct
  {
    char *in_filename;
    FILE *in;
    int pretty_print;
    int print_stats;
    int qdimacs_output;
    int deps_only;
    int print_deps;
    int dump_dep_graph;
    int trace;
    unsigned int max_time;
    unsigned int verbosity;
    unsigned int print_usage;
    unsigned int print_version;
    QDPLLA2AECCUcMode uc_mode;
    int uc_minimal;
    int uc_separate_csr;
    char *uc_filename;
    char *a2aecc_filename;
    FILE *a2aecc_file;
    int prefix_as_sets;
    int poldep_clauses;
    char *enuma2e_witness_basename;
  } options;
  char *comments;
  int max_len_comments;
  int cur_len_comments;
  VarID preamble_max_varid;
  unsigned int preamble_num_clauses;
  VarID prefixmatrix_max_varid;
  unsigned int prefixmatrix_num_clauses;
  unsigned int prefixmatrix_alt_depth;
  qdpll_a2aecc_in_occurs *in_occurs;
  ClauseGroupID *uc_prefix_clausegroupids, /* first element of array */
    *uc_prefix_clausegroupids_cand,        /* iterator */
    *uc_prefix_clausegroupids_next,        /* iterator, then after last assigned element of array */
    *uc_prefix_clausegroupids_end;         /* after last allocated element of array */
  ClauseGroupID *uc_matrix_clausegroupids, /* first element of array */
    *uc_matrix_clausegroupids_cand,        /* iterator */
    *uc_matrix_clausegroupids_next,        /* iterator, then after last assigned element of array */
    *uc_matrix_clausegroupids_end;         /* after last allocated element of array */
  VarID *uc_varid_to_a2aeccvarid,          /* first element of array */
    *uc_varid_to_a2aeccvarid_cand,         /* iterator */
    *uc_varid_to_a2aeccvarid_next,         /* iterator, then after last assigned element of array */
    *uc_varid_to_a2aeccvarid_end;          /* after last allocated element of array */
  int uc_atleast_one_clausegroup;
  unsigned int uc_successful_minimization_iterations;
  unsigned int uc_unsuccessful_minimization_iterations;
  unsigned int uc_unknown_minimization_iterations;
  VarID uc_max_varid;
  unsigned int uc_num_clauses;
  unsigned int uc_alt_depth;
  qdpll_a2aecc_core_occurs *core_occurs;
  QDPLLA2AECCVarOccCClauseGroupIDs *uc_varid_to_varocc_clausegroupids;
  unsigned int uc_enuma2e_irreducible;
  unsigned int uc_enuma2e_triviallyreducible;
  unsigned int uc_enuma2e_nontriviallyreducible;
  unsigned int uc_enuma2e_unknown;
};

typedef struct QDPLLA2AECCApp QDPLLA2AECCApp;

/* We keep a static pointer to the library object. Currently, this is
   used for calling library functions from within a signal handler. */
static QDPLL *qdpll = 0;

static void
print_abort_err (QDPLLA2AECCApp * app, char *msg, ...)
{
  va_list list;
  assert (msg != NULL);
  fprintf (stderr, "qdpll-a2aecc-app: ");
  va_start (list, msg);
  vfprintf (stderr, msg, list);
  va_end (list);
  fflush (stderr);
  abort ();
}


static void
qdpll_a2aecc_log (unsigned int level, QDPLLA2AECCApp * app, char *msg, ...)
{
  if (app->options.verbosity >= level) {
    va_list list;
    assert (msg != NULL);
    fprintf (stderr, "qdpll-a2aecc-app log: ");
    va_start (list, msg);
    vfprintf (stderr, msg, list);
    va_end (list);
    fflush (stderr);
  }
}


/* -------------------- START: PARSING -------------------- */
#define PARSER_READ_NUM(num, c)                        \
  assert (isdigit (c));                                \
  num = 0;					       \
  do						       \
    {						       \
      num = num * 10 + (c - '0');		       \
    }						       \
  while (isdigit ((c = getc (in))));

#define PARSER_SKIP_SPACE_DO_WHILE(c)		     \
  do						     \
    {                                                \
      c = getc (in);				     \
    }                                                \
  while (isspace (c));

#define PARSER_SKIP_SPACE_WHILE(c)		     \
  while (isspace (c))                                \
    c = getc (in);

#define PARSER_CHECK_LEN_COMMENTS                                       \
  do {                                                                  \
    if (app->cur_len_comments == app->max_len_comments) {               \
      app->comments = qdpll_realloc(mm,                                 \
                                    app->comments,                      \
                                    app->max_len_comments * sizeof(char), \
                                    (2 * app->max_len_comments + 1) * sizeof(char)); \
      app->max_len_comments = 2 * app->max_len_comments + 1;            \
    }                                                                   \
  } while (0)
  
#define PARSER_CHECK_LEN_UVTVC(puvtvc,mm)                         \
  do {                                                                  \
    if ((*(puvtvc)).varocc_clausegroupids == NULL) {                    \
      (*(puvtvc)).varocc_clausegroupids = (ClauseGroupID *)             \
        qdpll_malloc ((mm),                                             \
                      QDPLL_A2AECC_INIT_LEN_VAROCCCLAUSEGROUPIDS *      \
                      sizeof(ClauseGroupID));                           \
      (*(puvtvc)).varocc_clausegroupids_next = (*(puvtvc)).varocc_clausegroupids; \
      (*(puvtvc)).varocc_clausegroupids_end =                           \
        (*(puvtvc)).varocc_clausegroupids +                             \
        QDPLL_A2AECC_INIT_LEN_VAROCCCLAUSEGROUPIDS;                     \
    } else if ((*(puvtvc)).varocc_clausegroupids_next >=                \
               (*(puvtvc)).varocc_clausegroupids_end) {                 \
      unsigned int oldlen =                                             \
        (*(puvtvc)).varocc_clausegroupids_end -                         \
        (*(puvtvc)).varocc_clausegroupids;                              \
      (*(puvtvc)).varocc_clausegroupids =                               \
        qdpll_realloc((mm),                                             \
                      (*(puvtvc)).varocc_clausegroupids,                \
                      oldlen * sizeof(ClauseGroupID),                   \
                      (2 * oldlen + 1)  * sizeof(ClauseGroupID));       \
      (*(puvtvc)).varocc_clausegroupids_next =                          \
        (*(puvtvc)).varocc_clausegroupids + oldlen;                     \
      (*(puvtvc)).varocc_clausegroupids_end =                           \
        (*(puvtvc)).varocc_clausegroupids + 2 * oldlen + 1;             \
    }                                                                   \
  } while (0)


static void
flush_ascope_vars(QDPLLA2AECCApp * app,
                  QDPLL * qdpll,
                  int * a2aecc_existential_scope_open,
                  VarID * ascope_vars_begin,
                  VarID * ascope_vars_end)
{
  if (*a2aecc_existential_scope_open)
    qdpll_add (qdpll, 0);
  qdpll_new_scope (qdpll, QDPLL_QTYPE_FORALL);
  if (ascope_vars_begin == ascope_vars_end) {
    assert(0);
    QDPLL_A2AECC_ABORT_APP (1, "internal error: ascope_vars empty!\n");
    return;
  }
  for (VarID * ascope_vars_cand = ascope_vars_begin + 1;
       ascope_vars_cand < ascope_vars_end;
       ascope_vars_cand += 2) {
    if (ascope_vars_cand >= ascope_vars_end) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: end of ascope_vars reached!\n");
      return;
    }
    if (*ascope_vars_cand == 0) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: *ascope_vars_cand is 0!\n");
      return;
    }
    qdpll_add(qdpll, *ascope_vars_cand);
  }
  qdpll_add (qdpll, 0);
  qdpll_new_scope (qdpll, QDPLL_QTYPE_EXISTS);
  *a2aecc_existential_scope_open = 1;
  for (VarID * ascope_vars_cand = ascope_vars_begin;
       ascope_vars_cand < ascope_vars_end;
       ascope_vars_cand += 2) {
    if (ascope_vars_cand >= ascope_vars_end) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: end of ascope_vars reached!\n");
      return;
    }
    if (*ascope_vars_cand == 0) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: *ascope_vars_cand is 0!\n");
      return;
    }
    qdpll_add(qdpll, *ascope_vars_cand);
  }
}


static void
parse (QDPLLA2AECCApp * app, QDPLL * qdpll, FILE * in, int trace)
{
  QDPLLMemMan *mm = qdpll->mm;
  int c;
  int neg = 0;
  LitID num = 0;
  QDPLLQuantifierType scope_type = QDPLL_QTYPE_UNDEF;
  QDPLLQuantifierType first_nonempty_scope_type = QDPLL_QTYPE_UNDEF;
  QDPLLQuantifierType previous_scope_type = QDPLL_QTYPE_UNDEF;
  QDPLLQuantifierType previous_nonempty_scope_type = QDPLL_QTYPE_UNDEF;
  VarID a2aecc_varid;
  int scope_closed;
  int scope_empty;
  int a2aecc_existential_scope_open = 0;
  VarID * ascope_vars_begin = NULL,
    * ascope_vars_end = NULL;
  ClauseGroupID cgid = 0,
    cgid_todelete = 0;
  int clause_closed;
  int *uc_varid_occ = NULL;
  
  assert (in);

  if ((c = getc (in)) != EOF) {

    /* parse comments */
    if (!app->options.a2aecc_filename) {
      PARSER_SKIP_SPACE_WHILE (c);
      while (c == 'c') {
        while ((c = getc (in)) != '\n' && c != EOF)
          ;
        c = getc (in);
      }
    } else {
      app->comments = (char *)
        qdpll_malloc (mm, QDPLL_A2AECC_INIT_LEN_COMMENTS * sizeof(char));
      app->max_len_comments = QDPLL_A2AECC_INIT_LEN_COMMENTS;
      app->cur_len_comments = 0;
      PARSER_SKIP_SPACE_WHILE (c);
      while (c == 'c') {
        PARSER_CHECK_LEN_COMMENTS;
        app->comments[app->cur_len_comments++] = c;
        while ((c = getc (in)) != '\n' && c != EOF) {
          PARSER_CHECK_LEN_COMMENTS;
          app->comments[app->cur_len_comments++] = c;
        }
        if (c == '\n') {
          PARSER_CHECK_LEN_COMMENTS;
          app->comments[app->cur_len_comments++] = c;
        }
        c = getc (in);
      }
      PARSER_CHECK_LEN_COMMENTS;
      app->comments[app->cur_len_comments++] = 0;
    }

    /* parse preamble */
    PARSER_SKIP_SPACE_WHILE (c);
    if (c == 'p') {
      PARSER_SKIP_SPACE_DO_WHILE (c);
      if (c != 'c') {
        QDPLL_A2AECC_ABORT_APP (1, "malformed preamble!\n");
        return;
      }
      PARSER_SKIP_SPACE_DO_WHILE (c);
      if (c != 'n') {
        QDPLL_A2AECC_ABORT_APP (1, "malformed preamble!\n");
        return;
      }
      PARSER_SKIP_SPACE_DO_WHILE (c);
      if (c != 'f') {
        QDPLL_A2AECC_ABORT_APP (1, "malformed preamble!\n");
        return;
      }

      /* read number of variables */
      PARSER_SKIP_SPACE_DO_WHILE (c);
      if (!isdigit (c)) {
        QDPLL_A2AECC_ABORT_APP (1, "malformed preamble!\n");
        return;
      }
      PARSER_READ_NUM (app->preamble_max_varid, c);
      if (trace == TRACE_QRP)
        fprintf (stdout, "p qrp %u", app->preamble_max_varid);
      else if (trace == TRACE_BQRP)
        fprintf (stdout, "p bqrp %u", app->preamble_max_varid);

      /* read number of clauses */
      PARSER_SKIP_SPACE_WHILE (c);
      if (!isdigit (c)) {
        QDPLL_A2AECC_ABORT_APP (1, "malformed preamble!\n");
        return;
      }
      PARSER_READ_NUM (app->preamble_num_clauses, c);
      /* NOTE: number of steps is number of orig. clauses, as we can't tell
         the actual, future number of steps when tracing on-the-fly! */
      if (trace)
        fprintf (stdout,
                 " %u%c",
                 app->preamble_num_clauses,
                 trace == TRACE_QRP ? '\n' : 0);
    } else {
      QDPLL_A2AECC_ABORT_APP (1, "expecting preamble!\n");
      return;
    }

    /* set up data structures */
    app->in_occurs = (qdpll_a2aecc_in_occurs *)
      qdpll_malloc (mm, (app->preamble_max_varid + 1) * sizeof (qdpll_a2aecc_in_occurs));
    memset (app->in_occurs,
            0,
            (app->preamble_max_varid + 1) * sizeof (qdpll_a2aecc_in_occurs));
    if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2 ||
        app->options.a2aecc_filename) {
      /* uc_prefix_clausegroupids */
      app->uc_prefix_clausegroupids = (ClauseGroupID *)
        qdpll_malloc (mm, app->preamble_max_varid * sizeof (ClauseGroupID));
      memset (app->uc_prefix_clausegroupids,
              0,
              app->preamble_max_varid * sizeof (ClauseGroupID));
      app->uc_prefix_clausegroupids_next = app->uc_prefix_clausegroupids;
      app->uc_prefix_clausegroupids_end =
        app->uc_prefix_clausegroupids + app->preamble_max_varid;
      /* uc_varid_to_a2aeccvarid */
      app->uc_varid_to_a2aeccvarid = (VarID *)
        qdpll_malloc (mm, 2 * app->preamble_max_varid * sizeof (VarID));
      memset (app->uc_varid_to_a2aeccvarid,
              0,
              2 * app->preamble_max_varid * sizeof (VarID));
      app->uc_varid_to_a2aeccvarid_next = app->uc_varid_to_a2aeccvarid;
      app->uc_varid_to_a2aeccvarid_end =
        app->uc_varid_to_a2aeccvarid + 2 * app->preamble_max_varid;
      /* ascope_vars_begin,end */
      if (app->options.prefix_as_sets) {
        ascope_vars_begin = app->uc_varid_to_a2aeccvarid_next;
        ascope_vars_end = ascope_vars_begin;
      }
    }
    if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_C ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2) {
      /* uc_matrix_clausegroupids */
      app->uc_matrix_clausegroupids = (ClauseGroupID *)
        qdpll_malloc (mm, app->preamble_num_clauses * sizeof (ClauseGroupID));
      memset (app->uc_matrix_clausegroupids,
              0,
              app->preamble_num_clauses * sizeof (ClauseGroupID));
      app->uc_matrix_clausegroupids_next = app->uc_matrix_clausegroupids;
      app->uc_matrix_clausegroupids_end =
        app->uc_matrix_clausegroupids + app->preamble_num_clauses;
    }
    if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2) {
      /* uc_varid_occ */
      uc_varid_occ = (int *) qdpll_malloc (mm, (app->preamble_max_varid + 1) * sizeof (int));
      memset (uc_varid_occ,
              0,
              (app->preamble_max_varid + 1) * sizeof (int));
      /* uc_varocc_clausegroupids */
      app->uc_varid_to_varocc_clausegroupids = (QDPLLA2AECCVarOccCClauseGroupIDs *)
        qdpll_malloc (mm, (app->preamble_max_varid + 1) * sizeof (QDPLLA2AECCVarOccCClauseGroupIDs));
      memset (app->uc_varid_to_varocc_clausegroupids,
              0,
              (app->preamble_max_varid + 1) * sizeof (QDPLLA2AECCVarOccCClauseGroupIDs));
    }

    /* parse prefix */
    /* In uc modes q, qc, enuma2e1, enuma2e2 we maintain the invariant
       that there is an open existential scope. */
    a2aecc_varid = app->preamble_max_varid;
    PARSER_SKIP_SPACE_WHILE (c);
    while (c == 'a' || c == 'e') {
      if (c == 'a') {
        scope_type = QDPLL_QTYPE_FORALL;
        if (app->options.uc_mode != QDPLL_A2AECC_UC_MODE_Q &&
            app->options.uc_mode != QDPLL_A2AECC_UC_MODE_QC &&
            app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E1 &&
            app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E2 &&
            !app->options.a2aecc_filename) {
          qdpll_new_scope (qdpll, QDPLL_QTYPE_FORALL);
        }
      } else {
        scope_type = QDPLL_QTYPE_EXISTS;
        if ((app->options.uc_mode != QDPLL_A2AECC_UC_MODE_Q &&
             app->options.uc_mode != QDPLL_A2AECC_UC_MODE_QC &&
             app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E1 &&
             app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E2 &&
             !app->options.a2aecc_filename) ||
            !a2aecc_existential_scope_open) {
          qdpll_new_scope (qdpll, QDPLL_QTYPE_EXISTS);
          a2aecc_existential_scope_open = 1;
        }
      }
      if (scope_type == previous_scope_type) {
#if 0
        QDPLL_A2AECC_ABORT_APP (1, "malformed prefix!\n");
        return;
#endif
      } else {
        previous_scope_type = scope_type;
      }

      PARSER_SKIP_SPACE_DO_WHILE (c);
      scope_closed = 0;
      scope_empty = 1;
      while (!scope_closed) {
        if (!isdigit (c)) {
          QDPLL_A2AECC_ABORT_APP (1, "malformed prefix!\n");
          return;
        }
        PARSER_READ_NUM (num, c);
        if ((VarID) num > app->preamble_max_varid) {
          QDPLL_A2AECC_ABORT_APP (1, "varid larger than declared!\n");
          return;
        }
        if ((VarID) num > app->prefixmatrix_max_varid)
          app->prefixmatrix_max_varid = num;
        if (num == 0) {
          scope_closed = 1;
          if (app->options.uc_mode != QDPLL_A2AECC_UC_MODE_Q &&
              app->options.uc_mode != QDPLL_A2AECC_UC_MODE_QC &&
              app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E1 &&
              app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E2 &&
              !app->options.a2aecc_filename)
            qdpll_add (qdpll, 0);
        } else {
          if (app->in_occurs[num].prefix_a || app->in_occurs[num].prefix_e) {
            QDPLL_A2AECC_ABORT_APP (1, "malformed prefix!\n");
            return;
          }
          scope_empty = 0;
          if (first_nonempty_scope_type == QDPLL_QTYPE_UNDEF)
            first_nonempty_scope_type = scope_type;
          if (scope_type != previous_nonempty_scope_type) {
            if (previous_nonempty_scope_type == QDPLL_QTYPE_FORALL &&
                (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
                 app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC ||
                 app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
                 app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2 ||
                 app->options.a2aecc_filename) &&
                app->options.prefix_as_sets) {
              flush_ascope_vars(app,
                                qdpll,
                                &a2aecc_existential_scope_open,
                                ascope_vars_begin,
                                ascope_vars_end);
              ascope_vars_begin = app->uc_varid_to_a2aeccvarid_next;
              ascope_vars_end = ascope_vars_begin;
            }
            previous_nonempty_scope_type = scope_type;
            app->prefixmatrix_alt_depth++;
          }
          if (scope_type == QDPLL_QTYPE_FORALL) {
            if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
                app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC ||
                app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
                app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2 ||
                app->options.a2aecc_filename) {
              a2aecc_varid++;
              if (app->uc_varid_to_a2aeccvarid_next >=
                  app->uc_varid_to_a2aeccvarid_end) {
                assert(0);
                QDPLL_A2AECC_ABORT_APP (1, "internal error: no space left in uc_varid_to_a2aeccvarid!\n");
                return;
              } else {
                *app->uc_varid_to_a2aeccvarid_next = num;
                app->uc_varid_to_a2aeccvarid_next++;
                *app->uc_varid_to_a2aeccvarid_next = a2aecc_varid;
                app->uc_varid_to_a2aeccvarid_next++;
                if (app->options.prefix_as_sets)
                  ascope_vars_end += 2;
              }
              if (!app->options.prefix_as_sets) {
                if (a2aecc_existential_scope_open)
                  qdpll_add (qdpll, 0);
                qdpll_new_scope (qdpll, QDPLL_QTYPE_FORALL);
                qdpll_add (qdpll, a2aecc_varid);
                qdpll_add (qdpll, 0);
                qdpll_new_scope (qdpll, QDPLL_QTYPE_EXISTS);
                a2aecc_existential_scope_open = 1;
              }
            }
            app->in_occurs[num].prefix_a = 1;
          } else {
            app->in_occurs[num].prefix_e = 1;
          }
          if (scope_type != QDPLL_QTYPE_FORALL ||
              (app->options.uc_mode != QDPLL_A2AECC_UC_MODE_Q &&
               app->options.uc_mode != QDPLL_A2AECC_UC_MODE_QC &&
               app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E1 &&
               app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E2 &&
               !app->options.a2aecc_filename) ||
              !app->options.prefix_as_sets) {
            qdpll_add (qdpll, num);
          }
        }
        PARSER_SKIP_SPACE_WHILE (c);
      }
      if (scope_empty) {
#if 0
        QDPLL_A2AECC_ABORT_APP (1, "malformed prefix!\n");
        return;
#endif
      }
    }
    if ((app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
         app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC ||
         app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
         app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2 ||
         app->options.a2aecc_filename) &&
        app->options.prefix_as_sets &&
        ascope_vars_begin < ascope_vars_end) {
      flush_ascope_vars(app,
                        qdpll,
                        &a2aecc_existential_scope_open,
                        ascope_vars_begin,
                        ascope_vars_end);
      qdpll_add (qdpll, 0);
      a2aecc_existential_scope_open = 0;
    }
    /* If there was at least one scope, close it. */
    if ((app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
         app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC ||
         app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
         app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2 ||
         app->options.a2aecc_filename) &&
        a2aecc_existential_scope_open)
      qdpll_add (qdpll, 0);
    /* Consider possible undeclared existential outermost scope for
       app->prefixmatrix_alt_depth. Essentially
       qdpll.c::has_scope_free_user_var. */
    if (first_nonempty_scope_type == QDPLL_QTYPE_FORALL) {
      Scope *s = qdpll->pcnf.scopes.first;
      VarID *p, *e;
      for (p = s->vars.start, e = s->vars.top; p < e; p++) {
        VarID id = *p;
        Var *var = VARID2VARPTR(qdpll->pcnf.vars, id);
        assert (var->id == id);
        if (!var->is_internal &&
            !var->user_scope) {
          app->prefixmatrix_alt_depth++;
          break;
        }
      }
    }
                                             
    /* parse matrix */
    clause_closed = 1;
    if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_C ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2) {
      if (qdpll_get_open_clause_group(qdpll) != 0) {
        assert(0);
        QDPLL_A2AECC_ABORT_APP (1, "internal error: open clause group!\n");
        return;
      }
      cgid = qdpll_new_clause_group(qdpll);
      if (cgid == 0) {
        assert(0);
        QDPLL_A2AECC_ABORT_APP (1, "internal error: clause group id is 0!\n");
        return;
      }
      qdpll_open_clause_group(qdpll, cgid);
    }
    while (isdigit (c) || c == '-') {
      if (c == '-') {
        neg = 1;
        if (!isdigit ((c = getc (in)))) {
          QDPLL_A2AECC_ABORT_APP (1, "expecting digit!\n");
          return;
        }
      }

      /* parse a literal or variable ID */
      PARSER_READ_NUM (num, c);
      if ((VarID) num > app->preamble_max_varid) {
        QDPLL_A2AECC_ABORT_APP (1, "var id larger than declared!\n");
        return;
      }
      if ((VarID) num > app->prefixmatrix_max_varid)
        app->prefixmatrix_max_varid = num;
      if (num == 0) {
        clause_closed = 1;
        app->prefixmatrix_num_clauses++;
        if (app->prefixmatrix_num_clauses > app->preamble_num_clauses) {
          QDPLL_A2AECC_ABORT_APP (1, "number of clauses larger than declared!\n");
          return;
        }
        qdpll_add (qdpll, num);
        if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_C ||
            app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC ||
            app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
            app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2) {
          /* store cgid uc_varid_to_varocc */
          if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
              app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2) {
            for (VarID vid = 1; vid <= app->preamble_max_varid; vid++) {
              if (uc_varid_occ[vid]) {
                QDPLLA2AECCVarOccCClauseGroupIDs *puvtvc =
                  &app->uc_varid_to_varocc_clausegroupids[vid];
                PARSER_CHECK_LEN_UVTVC(puvtvc,mm);
                *puvtvc->varocc_clausegroupids_next = cgid;
                puvtvc->varocc_clausegroupids_next++;
              }
              uc_varid_occ[vid] = 0;
            }
          }
          /* close current clause group */
          if (qdpll_get_open_clause_group(qdpll) != cgid) {
            assert(0);
            QDPLL_A2AECC_ABORT_APP (1, "internal error: open clause group != cgid!\n");
            return;
          }
          qdpll_close_clause_group(qdpll, cgid);
          if (app->uc_matrix_clausegroupids_next >= app->uc_matrix_clausegroupids_end) {
            assert(0);
            QDPLL_A2AECC_ABORT_APP (1, "internal error: no space left in uc_matrix_clausegroupids!\n");
            return;
          } else {
            *app->uc_matrix_clausegroupids_next = cgid;
            app->uc_matrix_clausegroupids_next++;
          }
          /* open new clause group */
          cgid = qdpll_new_clause_group(qdpll);
          app->uc_atleast_one_clausegroup = 1;
          if (cgid == 0) {
            assert(0);
            QDPLL_A2AECC_ABORT_APP (1, "internal error: ClauseGroupID is 0!\n");
            return;
          }
          qdpll_open_clause_group(qdpll, cgid);
        }
      } else{
        clause_closed = 0;
        if (!neg)
          app->in_occurs[num].matrix_pos = 1;
        else
          app->in_occurs[num].matrix_neg = 1;
        if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
            app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2)
          uc_varid_occ[num] = 1;
        num = neg ? -num : num;
        qdpll_add (qdpll, num);
      }
      neg = 0;
      PARSER_SKIP_SPACE_WHILE (c);
    }
    if (!clause_closed) {
      QDPLL_A2AECC_ABORT_APP (1, "malformed matrix!\n");
      return;
    }
    if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_C ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2) {
      if (qdpll_get_open_clause_group(qdpll) != cgid) {
        assert(0);
        QDPLL_A2AECC_ABORT_APP (1, "internal error: open clause group != cgid!\n");
        return;
      }
      qdpll_close_clause_group(qdpll, cgid);
      cgid_todelete = cgid;
      /* We delete cgid_todelete after adding the a2aecc clauses. */
    }

    /* Add clauses (a2aecc_varid -> num), (!a2aecc_varid -> !num). */
    if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2 ||
        app->options.a2aecc_filename) {
      for (VarID *vidp = app->uc_varid_to_a2aeccvarid,
             *vide = app->uc_varid_to_a2aeccvarid_next;
           vidp < vide;
           vidp += 2) {
        VarID vid = *vidp;
        if (vid == 0) {
          assert(0);
          QDPLL_A2AECC_ABORT_APP (1, "internal error: vid is 0!\n");
          return;
        }
        if (app->in_occurs[vid].matrix_pos || app->in_occurs[vid].matrix_neg) {
          VarID vidaux = *(vidp+1);
          if (vidaux == 0) {
            assert(0);
            QDPLL_A2AECC_ABORT_APP (1, "internal error: vidaux is 0!\n");
            return;
          }
          if (qdpll_get_open_clause_group(qdpll) != 0) {
            assert(0);
            QDPLL_A2AECC_ABORT_APP (1, "internal error: open clause group!\n");
            return;
          }
          cgid = qdpll_new_clause_group(qdpll);
          app->uc_atleast_one_clausegroup = 1;
          if (cgid == 0) {
            assert(0);
            QDPLL_A2AECC_ABORT_APP (1, "internal error: ClauseGroupID is 0!\n");
            return;
          }
          qdpll_open_clause_group(qdpll, cgid);
          if (app->in_occurs[vid].matrix_neg || !app->options.poldep_clauses) {
            qdpll_add (qdpll, -vidaux);
            qdpll_add (qdpll, vid);
            qdpll_add (qdpll, 0);
          }
          if (app->in_occurs[vid].matrix_pos || !app->options.poldep_clauses) {
            qdpll_add (qdpll, vidaux);
            qdpll_add (qdpll, -vid);
            qdpll_add (qdpll, 0);
          }
          qdpll_close_clause_group(qdpll, cgid);
          if (app->uc_prefix_clausegroupids_next >=
              app->uc_prefix_clausegroupids_end) {
            assert(0);
            QDPLL_A2AECC_ABORT_APP (1, "internal error: no space left in uc_prefix_clausegroupids!\n");
            return;
          } else {
            *app->uc_prefix_clausegroupids_next = cgid;
            app->uc_prefix_clausegroupids_next++;
          }
          if (qdpll->options.verbosity >= 1) {
            if ((app->in_occurs[vid].matrix_pos &&
                 app->in_occurs[vid].matrix_neg) ||
                !app->options.poldep_clauses) {
            fprintf(stderr,
                    "converted 'a %d' into 'a %d e %d (%d, %d) (%d, %d)'\n",
                    vid,
                    vidaux,
                    vid,
                    -vidaux,
                    vid,
                    vidaux,
                    -vid);
            } else if (app->in_occurs[vid].matrix_neg) {
              fprintf(stderr,
                      "converted 'a %d' into 'a %d e %d (%d, %d)'\n",
                      vid,
                      vidaux,
                      vid,
                      -vidaux,
                      vid);
            } else {
              if (!app->in_occurs[vid].matrix_pos) {
                assert(0);
                QDPLL_A2AECC_ABORT_APP (1, "internal error: app->in_occurs[vid].matrix_pos is false!\n");
                return;
              }
              fprintf(stderr,
                      "converted 'a %d' into 'a %d e %d (%d, %d)'\n",
                      vid,
                      vidaux,
                      vid,
                      vidaux,
                      -vid);
            }
          }
        } else {
          /* Fill in dummy clause group to keep in sync with
             uc_varid_to_a2aeccvarid. */
          if (app->uc_prefix_clausegroupids_next >=
              app->uc_prefix_clausegroupids_end) {
            assert(0);
            QDPLL_A2AECC_ABORT_APP (1, "internal error: no space left in uc_prefix_clausegroupids!\n");
            return;
          } else {
            *app->uc_prefix_clausegroupids_next = 0;
            app->uc_prefix_clausegroupids_next++;
          }
        }
      }
    }

    /* We delete here rather than before adding a2aecc clauses to have
       the qdpll_gc after adding all clauses. */
    if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_C ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2) {
      if (!qdpll_exists_clause_group(qdpll, cgid_todelete)) {
        assert(0);
        QDPLL_A2AECC_ABORT_APP (1, "internal error: clause group does not exist!\n");
        return;
      }
      if (qdpll_get_open_clause_group(qdpll) == cgid_todelete) {
        assert(0);
        QDPLL_A2AECC_ABORT_APP (1, "internal error: open clause group == cgid_todelete!\n");
        return;
      }
      /* ideally we would assert that cgid is empty */
      qdpll_delete_clause_group(qdpll, cgid_todelete);
      qdpll_gc(qdpll);
    }

    /* parse EOF */
    if (c != EOF) {
      QDPLL_A2AECC_ABORT_APP (1, "expecting EOF!\n");
      return;
    }
#if 0
    if (app->prefixmatrix_max_varid != app->preamble_max_varid) {
      QDPLL_A2AECC_ABORT_APP (1, "actual maximum litid does not match declared maximum litid!\n");
      return;
    }
    if (app->prefixmatrix_num_clauses != app->preamble_num_clauses) {
      QDPLL_A2AECC_ABORT_APP (1, "actual number of clauses does not match declared number of clauses!\n");
      return;
    }
#endif
  } else {
    QDPLL_A2AECC_ABORT_APP (1, "preamble missing!\n");
    return;
  }

  /* cleanup */
  if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2) {
    if (uc_varid_occ) {
      qdpll_free(qdpll->mm,
                 uc_varid_occ,
                 (app->preamble_max_varid + 1) * sizeof (int));
    }
  }
  
  return;
}

/* -------------------- END: PARSING -------------------- */



static void
check_options (QDPLLA2AECCApp * app)
{
  if (app->options.print_deps && !app->options.deps_only)
    print_abort_err (app, "must not use option '%s' without option'%s'!\n\n",
                     "--print-deps", "--deps-only");

  if (app->options.dump_dep_graph && !app->options.deps_only)
    print_abort_err (app, "must not use option '%s' without option'%s'!\n\n",
                     "--dump-dep-graph", "--deps-only");

  if (app->options.uc_mode != QDPLL_A2AECC_UC_MODE_NONE &&
      !qdpll->options.incremental_use)
    print_abort_err (app, "must not use option '%s' without option'%s'!\n\n",
                     "--uc={c,q,qc,enuma2e1,enuma2e2}", "--incremental-use");

  if (app->options.uc_minimal &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_C &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_Q &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_QC)
    print_abort_err (app, "must not use option '%s' without option'%s'!\n\n",
                     "--min-uc", "--uc={c,q,qc}");

  if (app->options.uc_separate_csr &&
      !app->options.uc_minimal)
    print_abort_err (app, "must not use option '%s' without option'%s'!\n\n",
                     "--uc-separate-csr", "--uc=min-uc");

  if (app->options.uc_separate_csr &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_QC)
    print_abort_err (app, "must not use option '%s' without option'%s'!\n\n",
                     "--uc-separate-csr", "--uc=qc");

  if (app->options.uc_filename &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_C &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_Q &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_QC)
    print_abort_err (app, "must not use option '%s' without option'%s'!\n\n",
                     "--uc-file", "--uc={c,q,qc}");

  if (app->options.a2aecc_filename &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_NONE)
    print_abort_err (app, "must not use option '%s' with option '%s'!\n\n",
                     "--print-a2aecc", "--uc={c,q,qc,enuma2e1,enuma2e2}");
              
  if (app->options.prefix_as_sets &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_Q &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_QC &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E1 &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E2 &&
      !app->options.a2aecc_filename)
    print_abort_err (app, "must not use otpion '%s' without option '%s'!\n\n",
                     "--a2aecc-prefix-as-sets", "--uc={q,qc,enuma2e1,enuma2e2} or --print-a2aecc");
              
  if (app->options.poldep_clauses &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_Q &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_QC &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E1 &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E2 &&
      !app->options.a2aecc_filename)
    print_abort_err (app, "must not use otpion '%s' without option '%s'!\n\n",
                     "--a2aecc-poldep-clauses", "--uc={q,qc,enuma2e1,enuma2e2} or --print-a2aecc");
              
  if (app->options.enuma2e_witness_basename &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E1 &&
      app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E2)
    print_abort_err (app, "must not use option '%s' with option '%s'!\n\n",
                     "--enuma2e-witness-base", "--uc={enuma2e1,enuma2e2}");
}


static void
set_default_options (QDPLLA2AECCApp * app)
{
  app->options.in_filename = 0;
  app->options.in = stdin;
  app->options.pretty_print = 0;
  app->options.print_stats = 0;
  app->options.deps_only = 0;
  app->options.print_deps = 0;
  app->options.dump_dep_graph = 0;
  app->options.print_usage = 0;
  app->options.uc_mode = QDPLL_A2AECC_UC_MODE_NONE;
  app->options.uc_minimal = 0;
  app->options.uc_separate_csr = 0;
  app->options.uc_filename = 0;
  app->options.a2aecc_filename = 0;
  app->options.prefix_as_sets = 0;
  app->options.poldep_clauses = 0;
  app->options.enuma2e_witness_basename = 0;
}


static void
qdplla2aeccapp_init (QDPLLA2AECCApp * app)
{
  memset (app, 0, sizeof (QDPLLA2AECCApp));
  set_default_options(app);
}


static int
isnumstr (char *str)
{
  /* Empty string is not considered as number-string. */
  if (!*str)
    return 0;
  char *p;
  for (p = str; *p; p++)
    {
      if (!isdigit (*p))
        return 0;
    }
  return 1;
}


static void
parse_cmd_line_options (QDPLLA2AECCApp * app, QDPLL * qdpll, int argc, char **argv)
{
  char *result;
  int opt_cnt;
  for (opt_cnt = 1; opt_cnt < argc; opt_cnt++)
    {
      char *opt_str = argv[opt_cnt];

      if (!strcmp (opt_str, "-h") || !strcmp (opt_str, "--help"))
        {
          app->options.print_usage = 1;
        }
      else if (!strcmp (opt_str, "--version"))
        {
          app->options.print_version = 1;
        }
      else if (!strcmp (opt_str, "--pretty-print"))
        {
          app->options.pretty_print = 1;
        }
      else if (!strcmp (opt_str, "--print-stats"))
        {
          app->options.print_stats = 1;
        }
      else if (!strcmp (opt_str, "--qdo"))
        {
          app->options.qdimacs_output = 1;
        }
      else if (!strcmp (opt_str, "--deps-only"))
        {
          app->options.deps_only = 1;
        }
      else if (!strcmp (opt_str, "--print-deps"))
        {
          app->options.print_deps = 1;
        }
      else if (!strcmp (opt_str, "--dump-dep-graph"))
        {
          app->options.dump_dep_graph = 1;
        }
      else if (!strcmp (opt_str, "--qdag-print-deps-by-search"))
        {
          app->options.print_deps = 1;
          if ((result = qdpll_configure (qdpll, opt_str)))
            print_abort_err (app, "%s!\n\n", result);
        }
      else if (!strcmp (opt_str, "--trace")
               || !strcmp (opt_str, "--trace=qrp"))
        {
          app->options.trace = TRACE_QRP;
          qdpll_configure (qdpll, opt_str);
        }
      else if (!strcmp (opt_str, "--trace=bqrp"))
        {
          app->options.trace = TRACE_BQRP;
          qdpll_configure (qdpll, opt_str);
        }
      else if (!strcmp (opt_str, "--uc=none"))
        {
          if (app->options.uc_mode != QDPLL_A2AECC_UC_MODE_NONE)
            print_abort_err (app, "contradicting --uc options given!\n\n", result);
          app->options.uc_mode = QDPLL_A2AECC_UC_MODE_NONE;
        }
      else if (!strcmp (opt_str, "--uc=c"))
        {
          if (app->options.uc_mode != QDPLL_A2AECC_UC_MODE_NONE &&
              app->options.uc_mode != QDPLL_A2AECC_UC_MODE_C)
            print_abort_err (app, "contradicting --uc options given!\n\n", result);
          app->options.uc_mode = QDPLL_A2AECC_UC_MODE_C;
        }
      else if (!strcmp (opt_str, "--uc=q"))
        {
          if (app->options.uc_mode != QDPLL_A2AECC_UC_MODE_NONE &&
              app->options.uc_mode != QDPLL_A2AECC_UC_MODE_Q)
            print_abort_err (app, "contradicting --uc options given!\n\n", result);
          app->options.uc_mode = QDPLL_A2AECC_UC_MODE_Q;
        }
      else if (!strcmp (opt_str, "--uc=qc"))
        {
          if (app->options.uc_mode != QDPLL_A2AECC_UC_MODE_NONE &&
              app->options.uc_mode != QDPLL_A2AECC_UC_MODE_QC)
            print_abort_err (app, "contradicting --uc options given!\n\n", result);
          app->options.uc_mode = QDPLL_A2AECC_UC_MODE_QC;
        }
      else if (!strcmp (opt_str, "--uc=enuma2e1"))
        {
          if (app->options.uc_mode != QDPLL_A2AECC_UC_MODE_NONE &&
              app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E1)
            print_abort_err (app, "contradicting --uc options given!\n\n", result);
          app->options.uc_mode = QDPLL_A2AECC_UC_MODE_ENUMA2E1;
        }
      else if (!strcmp (opt_str, "--uc=enuma2e2"))
        {
          if (app->options.uc_mode != QDPLL_A2AECC_UC_MODE_NONE &&
              app->options.uc_mode != QDPLL_A2AECC_UC_MODE_ENUMA2E2)
            print_abort_err (app, "contradicting --uc options given!\n\n", result);
          app->options.uc_mode = QDPLL_A2AECC_UC_MODE_ENUMA2E2;
        }
      else if (!strcmp (opt_str, "--min-uc"))
        {
          app->options.uc_minimal = 1;
        }
      else if (!strcmp (opt_str, "--uc-separate-csr"))
        {
          app->options.uc_separate_csr = 1;
        }
      else if (!strcmp (opt_str, "--uc-file"))
        {
          opt_cnt++;
          if (opt_cnt == argc)
            print_abort_err (app, "option --uc-file requires file name argument!\n\n");
          opt_str = argv[opt_cnt];
          if (app->options.uc_filename)
            print_abort_err (app, "duplicate occurrence of option --uc-file!\n\n");
          app->options.uc_filename = opt_str;
          /* Check uc file. */
          DIR *dir;
          if ((dir = opendir (app->options.uc_filename)) != NULL)
            {
              closedir (dir);
              print_abort_err (app, "uc file '%s' is a directory!\n\n",
                               app->options.uc_filename);
            }
        }
      else if (!strcmp (opt_str, "--print-a2aecc"))
        {
          opt_cnt++;
          if (opt_cnt == argc)
            print_abort_err (app, "option --print-a2aecc requires file name argument!\n\n");
          opt_str = argv[opt_cnt];
          if (app->options.a2aecc_filename)
            print_abort_err (app, "duplicate occurrence of option --print-a2aecc!\n\n");
          app->options.a2aecc_filename = opt_str;
          /* Check a2aecc file. */
          DIR *dir;
          if ((dir = opendir (app->options.a2aecc_filename)) != NULL)
            {
              closedir (dir);
              print_abort_err (app, "a2aecc file '%s' is a directory!\n\n",
                               app->options.a2aecc_filename);
            }
          FILE *a2aecc_file = fopen (app->options.a2aecc_filename, "w");
          if (!a2aecc_file)
            {
              print_abort_err (app, "could not open a2aecc file '%s'!\n\n",
                               app->options.a2aecc_filename);
            }
          else
            {
              app->options.a2aecc_file = a2aecc_file;
              if ((result = qdpll_configure (qdpll, "--dep-man=simple")))
                print_abort_err (app, "%s!\n\n", result);
              if ((result = qdpll_configure (qdpll, "--incremental-use")))
                print_abort_err (app, "%s!\n\n", result);
            }
        }
      else if (!strcmp (opt_str, "--a2aecc-prefix-as-sets"))
        {
          app->options.prefix_as_sets = 1;
        }
      else if (!strcmp (opt_str, "--a2aecc-poldep-clauses"))
        {
          app->options.poldep_clauses = 1;
        }
      else if (!strcmp (opt_str, "--enuma2e-witness-base"))
        {
          opt_cnt++;
          if (opt_cnt == argc)
            print_abort_err (app, "option --enuma2e-witness-base requires argument!\n\n");
          opt_str = argv[opt_cnt];
          if (app->options.enuma2e_witness_basename)
            print_abort_err (app, "duplicate occurrence of option --enuma2e-witness-base!\n\n");
          app->options.enuma2e_witness_basename = opt_str;
        }
      else if (isnumstr (opt_str))
        {
          app->options.max_time = atoi (opt_str);
          if (app->options.max_time == 0)
            {
              result = "Expecting non-zero value for max-time";
              print_abort_err (app, "%s!\n\n", result);
            }
        }
      else if (!strncmp (opt_str, "-", 1) || !strncmp (opt_str, "--", 2))
        {
          /* Handle QDPLL-options. */
          if (!strcmp (opt_str, "-v"))
            {
              app->options.verbosity++;
            }
          if ((result = qdpll_configure (qdpll, opt_str)))
            print_abort_err (app, "%s!\n\n", result);
        }
      else if (!app->options.in_filename)
        {
          app->options.in_filename = opt_str;
          /* Check input file. */
          DIR *dir;
          if ((dir = opendir (app->options.in_filename)) != NULL)
            {
              closedir (dir);
              print_abort_err (app, "input file '%s' is a directory!\n\n",
                               app->options.in_filename);
            }
          FILE *input_file = fopen (app->options.in_filename, "r");
          if (!input_file)
            {
              print_abort_err (app, "could not open input file '%s'!\n\n",
                               app->options.in_filename);
            }
          else
            app->options.in = input_file;
        }
      else
        {
          print_abort_err (app, "unknown option '%s'!\n\n", opt_str);
        }
    }
}


static void
print_deps (QDPLLA2AECCApp * app, QDPLL * qdpll)
{
  VarID i, max;
  for (i = 1, max = qdpll_get_max_declared_var_id (qdpll); i <= max; i++)
    {
      fprintf (stdout, "%u: ", i);
      qdpll_print_deps (qdpll, i);
      fprintf (stdout, "\n");
    }
  fprintf (stdout, "\n");
}


static void
sig_handler (int sig)
{
  fprintf (stderr, "\n\n SIG RECEIVED\n\n");
#if (COMPUTE_STATS || COMPUTE_TIMES)
  qdpll_print_stats (qdpll);
#endif
  signal (sig, SIG_DFL);
  raise (sig);
}


static void
sigalrm_handler (int sig)
{
  fprintf (stderr, "\n\n SIGALRM RECEIVED\n\n");
#if (COMPUTE_STATS || COMPUTE_TIMES)
  qdpll_print_stats (qdpll);
#endif
  signal (sig, SIG_DFL);
  raise (sig);
}


static void
set_signal_handlers (QDPLLA2AECCApp * app)
{
  signal (SIGINT, sig_handler);
  signal (SIGTERM, sig_handler);
  signal (SIGALRM, sigalrm_handler);
  signal (SIGXCPU, sigalrm_handler);
}


static cgidhash_t*
fill_cgidh_with_relevant_clausegroups (ClauseGroupID *relevant_cgids)
{
  ClauseGroupID *relevant_cgids_cand;
  cgidhash_t *cgidh;

  for (relevant_cgids_cand = relevant_cgids;
       *relevant_cgids_cand;
       relevant_cgids_cand++)
    /* NOP */;
  cgidh = cgidhash_new(qdpll->mm, 2 * (relevant_cgids_cand - relevant_cgids));
  for (relevant_cgids_cand = relevant_cgids;
       *relevant_cgids_cand;
       relevant_cgids_cand++) {
    if (*relevant_cgids_cand == 0) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: clause group id is 0!\n");
      return NULL;
    }
    cgidhash_insert(cgidh, *relevant_cgids_cand);
  }
  return cgidh;
}


static void
delete_nonoccurring_clausegroups (QDPLLA2AECCApp * app,
                                  QDPLL *qdpll,
                                  ClauseGroupID *beg,
                                  ClauseGroupID *end,
                                  cgidhash_t *cgidh)
{
  unsigned int num_clausegroups_deleted = 0;
  
  for (ClauseGroupID *cgid = beg;
       cgid < end;
       cgid++) {
    if (*cgid != 0) {
      if (!cgidhash_present(cgidh, *cgid)) {
        if (!qdpll_exists_clause_group(qdpll, *cgid)) {
          assert(0);
          QDPLL_A2AECC_ABORT_APP (1, "internal error: clause group does not exist!\n");
          return;
        }
        if (!qdpll_isactive_clause_group(qdpll, *cgid)) {
          assert(0);
          QDPLL_A2AECC_ABORT_APP (1, "internal error: clause group is not active!\n");
          return;
        }
        num_clausegroups_deleted++;
        qdpll_delete_clause_group(qdpll, *cgid);
        /* We do qdpll_gc in delete_irrelevant_clausegroups. */
        *cgid = 0;
      }
    }
  }
  qdpll_a2aecc_log(1, app, "deleted %d clause groups\n", num_clausegroups_deleted);
}


static void
delete_irrelevant_clausegroups (QDPLLA2AECCApp * app,
                                QDPLL * qdpll,
                                ClauseGroupID toactivate,
                                int delete_q_only)
{
  ClauseGroupID *relevant_cgids;
  cgidhash_t *cgidh;

  assert(!delete_q_only ||
         (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC &&
          app->options.uc_minimal &&
          app->options.uc_separate_csr));
  
  relevant_cgids = qdpll_get_relevant_clause_groups(qdpll);
  qdpll_reset(qdpll);
  if (toactivate)
    qdpll_activate_clause_group(qdpll, toactivate);
  
  cgidh = fill_cgidh_with_relevant_clausegroups(relevant_cgids);
  if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC) {
    delete_nonoccurring_clausegroups(app,
                                     qdpll,
                                     app->uc_prefix_clausegroupids_cand,
                                     app->uc_prefix_clausegroupids_next,
                                     cgidh);
  }
  if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_C ||
      (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC &&
       !delete_q_only)) {
    delete_nonoccurring_clausegroups(app,
                                     qdpll,
                                     app->uc_matrix_clausegroupids_cand,
                                     app->uc_matrix_clausegroupids_next,
                                     cgidh);
  }
  cgidhash_delete(qdpll->mm, cgidh);
  free(relevant_cgids);
  qdpll_gc (qdpll);
}


static void
minimize_clausegroups (QDPLLA2AECCApp * app,
                       QDPLL * qdpll,
                       ClauseGroupID *cand,
                       ClauseGroupID *end)
{
  QDPLLResult tmpres;
  
  for (;
       cand < end;
       cand++) {
    if (*cand != 0) {
      if (!qdpll_exists_clause_group(qdpll, *cand)) {
        assert(0);
        QDPLL_A2AECC_ABORT_APP (1, "internal error: clause group does not exist!\n");
        return;
      }
      qdpll_deactivate_clause_group(qdpll, *cand);
      qdpll_a2aecc_log(1, app, "begin checking satisfiability of weakened formula\n");
      tmpres = qdpll_sat(qdpll);
      qdpll_a2aecc_log(1, app, "end checking satisfiability of weakened formula\n");
      if (tmpres == QDPLL_RESULT_UNSAT) {
        app->uc_successful_minimization_iterations++;
        qdpll_a2aecc_log(1, app, "begin deleting irrelevant clause groups\n");
        if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC &&
            app->options.uc_separate_csr &&
            end == app->uc_prefix_clausegroupids_next /* we're minimizing the quantifiers */)
          delete_irrelevant_clausegroups(app, qdpll, *cand, 1);
        else
          delete_irrelevant_clausegroups(app, qdpll, *cand, 0);
        qdpll_a2aecc_log(1, app, "end deleting irrelevant clause groups\n");
        /* We did qdpll_reset in delete_irrelevant_clausegroups after
           qdpll_get_relevant_clause_groups. */
      } else {
        if (tmpres == QDPLL_RESULT_SAT) {
          app->uc_unsuccessful_minimization_iterations++;
        } else {
          app->uc_unknown_minimization_iterations++;
        }
        qdpll_reset(qdpll);
        qdpll_activate_clause_group(qdpll, *cand);
      }
    }
  }
}


static void
print_statistics (QDPLLA2AECCApp * app, FILE * stream)
{
  unsigned int in_a_both = 0;  /* prefix: a; matrix: +,- */
  unsigned int in_a_pos = 0;   /* prefix: a; matrix: +   */
  unsigned int in_a_neg = 0;   /* prefix: a; matrix: -   */
  unsigned int in_a_0 = 0;     /* prefix: a; matrix: 0   */
  unsigned int in_e_both = 0;  /* prefix: e; matrix: +,- */
  unsigned int in_e_pos = 0;   /* prefix: e; matrix: +   */
  unsigned int in_e_neg = 0;   /* prefix: e; matrix: -   */
  unsigned int in_e_0 = 0;     /* prefix: e; matrix: 0   */
  unsigned int in_ei_both = 0; /* prefix: 0; matrix: +,- */
  unsigned int in_ei_pos = 0;  /* prefix: 0; matrix: +   */
  unsigned int in_ei_neg = 0;  /* prefix: 0; matrix: -   */
  unsigned int in_0 = 0;       /* prefix: 0; matrix: 0   */ /* Those are gaps in the range of available variable ids. */

  unsigned int uc_a_both = 0;  /* prefix: a; matrix: +,- */
  unsigned int uc_a_pos = 0;   /* prefix: a; matrix: +   */
  unsigned int uc_a_neg = 0;   /* prefix: a; matrix: -   */
  unsigned int uc_e_both = 0;  /* prefix: e; matrix: +,- */
  unsigned int uc_e_pos = 0;   /* prefix: e; matrix: +   */
  unsigned int uc_e_neg = 0;   /* prefix: e; matrix: -   */
  unsigned int uc_0 = 0;       /* prefix: 0; matrix: 0   */ /* Those are gaps in the range of available variable ids. */

  unsigned int in_a_both_uc_a_both = 0;
  unsigned int in_a_both_uc_a_pos = 0;
  unsigned int in_a_both_uc_a_neg = 0;
  unsigned int in_a_both_uc_e_both = 0;
  unsigned int in_a_both_uc_e_pos = 0;
  unsigned int in_a_both_uc_e_neg = 0;
  unsigned int in_a_both_uc_0 = 0;
  unsigned int in_a_pos_uc_a_pos = 0;
  unsigned int in_a_pos_uc_e_pos = 0;
  unsigned int in_a_pos_uc_0 = 0;
  unsigned int in_a_neg_uc_a_neg = 0;
  unsigned int in_a_neg_uc_e_neg = 0;
  unsigned int in_a_neg_uc_0 = 0;
  unsigned int in_e_both_uc_e_both = 0;
  unsigned int in_e_both_uc_e_pos = 0;
  unsigned int in_e_both_uc_e_neg = 0;
  unsigned int in_e_both_uc_0 = 0;
  unsigned int in_e_pos_uc_e_pos = 0;
  unsigned int in_e_pos_uc_0 = 0;
  unsigned int in_e_neg_uc_e_neg = 0;
  unsigned int in_e_neg_uc_0 = 0;

  VarID vid;

  for (vid = 1; vid <= app->preamble_max_varid; vid++) {
    qdpll_a2aecc_in_occurs io = app->in_occurs[vid];
    if (io.prefix_a && io.prefix_e) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: vid has true in_occurs.prefix_a and in_occurs.prefix_e!\n");
      return;
    }
    if (io.prefix_a) {
      if (io.matrix_pos && io.matrix_neg)
        in_a_both++;
      else if (io.matrix_pos && !io.matrix_neg)
        in_a_pos++;
      else if (!io.matrix_pos && io.matrix_neg)
        in_a_neg++;
      else
        in_a_0++;
    } else if (io.prefix_e) {
      if (io.matrix_pos && io.matrix_neg)
        in_e_both++;
      else if (io.matrix_pos && !io.matrix_neg)
        in_e_pos++;
      else if (!io.matrix_pos && io.matrix_neg)
        in_e_neg++;
      else
        in_e_0++;
    } else {
      if (io.matrix_pos && io.matrix_neg)
        in_ei_both++;
      else if (io.matrix_pos && !io.matrix_neg)
        in_ei_pos++;
      else if (!io.matrix_pos && io.matrix_neg)
        in_ei_neg++;
      else
        in_0++;
    }
  }
  if (in_a_both +
      in_a_pos +
      in_a_neg +
      in_a_0 +
      in_e_both +
      in_e_pos +
      in_e_neg +
      in_e_0 +
      in_ei_both +
      in_ei_pos +
      in_ei_neg +
      in_0
      !=
      app->preamble_max_varid) {
    assert(0);
    QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
    return;
  }
  
  if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_C ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC) {
    for (vid = 1; vid <= app->uc_max_varid; vid++) {
      qdpll_a2aecc_in_occurs io = app->in_occurs[vid];
      qdpll_a2aecc_core_occurs uo = app->core_occurs[vid];
      if ((app->options.uc_mode == QDPLL_A2AECC_UC_MODE_C && io.prefix_a) ||
          ((app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
            app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC) && uo.matrix_a2aecc)) {
        if (uo.matrix_pos && uo.matrix_neg)
          uc_a_both++;
        else if (uo.matrix_pos && !uo.matrix_neg)
          uc_a_pos++;
        else if (!uo.matrix_pos && uo.matrix_neg)
          uc_a_neg++;
        else {
          uc_0++;
        }
      } else {
        if (uo.matrix_pos && uo.matrix_neg)
          uc_e_both++;
        else if (uo.matrix_pos && !uo.matrix_neg)
          uc_e_pos++;
        else if (!uo.matrix_pos && uo.matrix_neg)
          uc_e_neg++;
        else
          uc_0++;
      }
    }
    if (uc_a_both +
        uc_a_pos +
        uc_a_neg +
        uc_e_both +
        uc_e_pos +
        uc_e_neg +
        uc_0
        !=
        app->uc_max_varid) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
      return;
    }

    for (vid = 1; vid <= app->prefixmatrix_max_varid; vid++) {
      qdpll_a2aecc_in_occurs io = app->in_occurs[vid];
      if (vid <= app->uc_max_varid) {
        qdpll_a2aecc_core_occurs uo = app->core_occurs[vid];
        if (io.prefix_a) {
          if (io.matrix_pos && io.matrix_neg) {
            if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_C ||
                ((app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
                  app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC) && uo.matrix_a2aecc)) {
              if (uo.matrix_pos && uo.matrix_neg)
                in_a_both_uc_a_both++;
              else if (uo.matrix_pos && !uo.matrix_neg)
                in_a_both_uc_a_pos++;
              else if (!uo.matrix_pos && uo.matrix_neg)
                in_a_both_uc_a_neg++;
              else {
                in_a_both_uc_0++;
              }
            } else {
              if (uo.matrix_pos && uo.matrix_neg)
                in_a_both_uc_e_both++;
              else if (uo.matrix_pos && !uo.matrix_neg)
                in_a_both_uc_e_pos++;
              else if (!uo.matrix_pos && uo.matrix_neg)
                in_a_both_uc_e_neg++;
              else
                in_a_both_uc_0++;
            }
          } else if (io.matrix_pos && !io.matrix_neg) {
            assert(!uo.matrix_neg);
            if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_C ||
                ((app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
                app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC) && uo.matrix_a2aecc)) {
              if (uo.matrix_pos)
                in_a_pos_uc_a_pos++;
              else {
                in_a_pos_uc_0++;
              }
            } else {
              if (uo.matrix_pos)
                in_a_pos_uc_e_pos++;
              else
                in_a_pos_uc_0++;
            }
          } else if (!io.matrix_pos && io.matrix_neg) {
            assert(!uo.matrix_pos);
            if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_C ||
                ((app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
                  app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC) && uo.matrix_a2aecc)) {
              if (uo.matrix_neg)
                in_a_neg_uc_a_neg++;
              else {
                in_a_neg_uc_0++;
              }
            } else {
              if (uo.matrix_neg)
                in_a_neg_uc_e_neg++;
              else
                in_a_neg_uc_0++;
            }
          }
        } else if (io.prefix_e) {
          assert(!uo.matrix_a2aecc);
          if (io.matrix_pos && io.matrix_neg) {
            if (uo.matrix_pos && uo.matrix_neg)
              in_e_both_uc_e_both++;
            else if (uo.matrix_pos && !uo.matrix_neg)
              in_e_both_uc_e_pos++;
            else if (!uo.matrix_pos && uo.matrix_neg)
              in_e_both_uc_e_neg++;
            else
              in_e_both_uc_0++;
          } else if (io.matrix_pos && !io.matrix_neg) {
            assert(!uo.matrix_neg);
            if (uo.matrix_pos)
              in_e_pos_uc_e_pos++;
            else
              in_e_pos_uc_0++;
          } else if (!io.matrix_pos && io.matrix_neg) {
            assert(!uo.matrix_pos);
            if (uo.matrix_neg)
              in_e_neg_uc_e_neg++;
            else
              in_e_neg_uc_0++;
          }
        }
      } else {
        if (io.prefix_a) {
          if (io.matrix_pos && io.matrix_neg) {
            in_a_both_uc_0++;
          } else if (io.matrix_pos && !io.matrix_neg) {
            in_a_pos_uc_0++;
          } else if (!io.matrix_pos && io.matrix_neg) {
            in_a_neg_uc_0++;
          }
        } else if (io.prefix_e) {
          if (io.matrix_pos && io.matrix_neg) {
            in_e_both_uc_0++;
          } else if (io.matrix_pos && !io.matrix_neg) {
            in_e_pos_uc_0++;
          } else if (!io.matrix_pos && io.matrix_neg) {
            in_e_neg_uc_0++;
          }
        }
      }
    }
    if (in_a_both
        !=
        in_a_both_uc_a_both +
        in_a_both_uc_a_pos +
        in_a_both_uc_a_neg +
        in_a_both_uc_e_both +
        in_a_both_uc_e_pos +
        in_a_both_uc_e_neg +
        in_a_both_uc_0) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
      return;
    }
    if (in_a_pos
        !=
        in_a_pos_uc_a_pos +
        in_a_pos_uc_e_pos +
        in_a_pos_uc_0) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
      return;
    }
    if (in_a_neg
        !=
        in_a_neg_uc_a_neg +
        in_a_neg_uc_e_neg +
        in_a_neg_uc_0) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
      return;
    }
    if (in_e_both +
        in_ei_both
        !=
        in_e_both_uc_e_both +
        in_e_both_uc_e_pos +
        in_e_both_uc_e_neg +
        in_e_both_uc_0) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
      return;
    }
    if (in_e_pos +
        in_ei_pos
        !=
        in_e_pos_uc_e_pos +
        in_e_pos_uc_0) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
      return;
    }
    if (in_e_neg +
        in_ei_neg
        !=
        in_e_neg_uc_e_neg +
        in_e_neg_uc_0) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
      return;
    }
    if (uc_a_both
        !=
        in_a_both_uc_a_both) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
      return;
    }
    if (uc_a_pos
        !=
        in_a_both_uc_a_pos +
        in_a_pos_uc_a_pos) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
      return;
    }
    if (uc_a_neg
        !=
        in_a_both_uc_a_neg +
        in_a_neg_uc_a_neg) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
      return;
    }
    if (uc_e_both
        !=
        in_a_both_uc_e_both +
        in_e_both_uc_e_both) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
      return;
    }
    if (uc_e_pos
        !=
        in_a_both_uc_e_pos +
        in_a_pos_uc_e_pos +
        in_e_both_uc_e_pos +
        in_e_pos_uc_e_pos) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
      return;
    }
    if (uc_e_neg
        !=
        in_a_both_uc_e_neg +
        in_a_neg_uc_e_neg +
        in_e_both_uc_e_neg +
        in_e_neg_uc_e_neg) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
      return;
    }
    if (uc_0 +
        (app->preamble_max_varid - app->uc_max_varid)
        !=
        in_a_both_uc_0 +
        in_a_pos_uc_0 +
        in_a_neg_uc_0 +
        in_a_0 +
        in_e_both_uc_0 +
        in_e_pos_uc_0 +
        in_e_neg_uc_0 +
        in_e_0 +
        in_0) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
      return;
    }
  }

  if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2) {
    if (app->uc_enuma2e_irreducible +
        app->uc_enuma2e_nontriviallyreducible +
        app->uc_enuma2e_triviallyreducible +
        app->uc_enuma2e_unknown
        !=
        in_a_both +
        in_a_pos +
        in_a_neg +
        in_a_0) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: wrong statistics!\n");
      return;
    }
  }

  fprintf(stream, "Preamble:\n");
  fprintf(stream, "  P maximum variable id:                       %u\n", app->preamble_max_varid);
  fprintf(stream, "  P number of clauses:                         %u\n", app->preamble_num_clauses);
  fprintf(stream, "Input formula:\n");
  fprintf(stream, "  IF maximum variable id:                      %u\n", app->prefixmatrix_max_varid);
  fprintf(stream, "  IF number of clauses:                        %u\n", app->prefixmatrix_num_clauses);
  fprintf(stream, "  IF alternation depth:                        %u\n", app->prefixmatrix_alt_depth);
  fprintf(stream, "  IF A quantified, occurs both:                %u\n", in_a_both);
  fprintf(stream, "  IF A quantified, occurs positive:            %u\n", in_a_pos);
  fprintf(stream, "  IF A quantified, occurs negative:            %u\n", in_a_neg);
  fprintf(stream, "  IF A quantified, no occurrences:             %u\n", in_a_0);
  fprintf(stream, "  IF explicitly E quantified, occurs both:     %u\n", in_e_both);
  fprintf(stream, "  IF explicitly E quantified, occurs positive: %u\n", in_e_pos);
  fprintf(stream, "  IF explicitly E quantified, occurs negative: %u\n", in_e_neg);
  fprintf(stream, "  IF explicitly E quantified, no occurrences:  %u\n", in_e_0);
  fprintf(stream, "  IF implicitly E quantified, occurs both:     %u\n", in_ei_both);
  fprintf(stream, "  IF implicitly E quantified, occurs positive: %u\n", in_ei_pos);
  fprintf(stream, "  IF implicitly E quantified, occurs negative: %u\n", in_ei_neg);
  fprintf(stream, "  IF no occurrences (gaps in legal range):     %u\n", in_0);
  if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_C ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC) {
    fprintf(stream, "Unsatisfiable core:\n");
    fprintf(stream, "  UC maximum variable id:                      %u\n", app->uc_max_varid);
    fprintf(stream, "  UC number of clauses:                        %u\n", app->uc_num_clauses);
    fprintf(stream, "  UC alternation depth:                        %u\n", app->uc_alt_depth);
    fprintf(stream, "  UC A quantified, occurs both:                %u\n", uc_a_both);
    fprintf(stream, "  UC A quantified, occurs positive:            %u\n", uc_a_pos);
    fprintf(stream, "  UC A quantified, occurs negative:            %u\n", uc_a_neg);
    fprintf(stream, "  UC E quantified, occurs both:                %u\n", uc_e_both);
    fprintf(stream, "  UC E quantified, occurs positive:            %u\n", uc_e_pos);
    fprintf(stream, "  UC E quantified, occurs negative:            %u\n", uc_e_neg);
    fprintf(stream, "  UC no occurrences (gaps in legal range):     %u\n", uc_0);
    fprintf(stream, "Changes input formula to unsatisfiable core:\n");
    fprintf(stream, "  IF2UC removed variable ids:                  %u\n",
            (app->preamble_max_varid - in_0) - (app->uc_max_varid - uc_0));
    fprintf(stream, "  IF2UC removed clauses:                       %u\n",
            app->prefixmatrix_num_clauses - app->uc_num_clauses);
    fprintf(stream, "  IF2UC reduction in alternation depth:        %d\n",
            app->prefixmatrix_alt_depth - app->uc_alt_depth);
    fprintf(stream, "  IF2UC weakened A quantified to E quantified: %u\n",
            in_a_both_uc_e_both +
            in_a_both_uc_e_pos +
            in_a_both_uc_e_neg +
            in_a_pos_uc_e_pos +
            in_a_neg_uc_e_neg);
    fprintf(stream, "  IF2UC A both to A both:                      %u\n", in_a_both_uc_a_both);
    fprintf(stream, "  IF2UC A both to A positive:                  %u\n", in_a_both_uc_a_pos);
    fprintf(stream, "  IF2UC A both to A negative:                  %u\n", in_a_both_uc_a_neg);
    fprintf(stream, "  IF2UC A both to E both:                      %u\n", in_a_both_uc_e_both);
    fprintf(stream, "  IF2UC A both to E positive:                  %u\n", in_a_both_uc_e_pos);
    fprintf(stream, "  IF2UC A both to E negative:                  %u\n", in_a_both_uc_e_neg);
    fprintf(stream, "  IF2UC A both to no occurrence:               %u\n", in_a_both_uc_0);
    fprintf(stream, "  IF2UC A positive to A positive:              %u\n", in_a_pos_uc_a_pos);
    fprintf(stream, "  IF2UC A positive to E positive:              %u\n", in_a_pos_uc_e_pos);
    fprintf(stream, "  IF2UC A positive to no occurrence:           %u\n", in_a_pos_uc_0);
    fprintf(stream, "  IF2UC A negative to A negative:              %u\n", in_a_neg_uc_a_neg);
    fprintf(stream, "  IF2UC A negative to E negative:              %u\n", in_a_neg_uc_e_neg);
    fprintf(stream, "  IF2UC A negative to no occurrence:           %u\n", in_a_neg_uc_0);
    fprintf(stream, "  IF2UC E both to E both:                      %u\n", in_e_both_uc_e_both);
    fprintf(stream, "  IF2UC E both to E positive:                  %u\n", in_e_both_uc_e_pos);
    fprintf(stream, "  IF2UC E both to E negative:                  %u\n", in_e_both_uc_e_neg);
    fprintf(stream, "  IF2UC E both to no occurrence:               %u\n", in_e_both_uc_0);
    fprintf(stream, "  IF2UC E positive to E positive:              %u\n", in_e_pos_uc_e_pos);
    fprintf(stream, "  IF2UC E positive to no occurrence:           %u\n", in_e_pos_uc_0);
    fprintf(stream, "  IF2UC E negative to E negative:              %u\n", in_e_neg_uc_e_neg);
    fprintf(stream, "  IF2UC E negative to no occurrence:           %u\n", in_e_neg_uc_0);
    if (app->options.uc_minimal) {
      fprintf(stream, "Minimization of unsatisfiable core:\n");
      fprintf(stream, "  Iterations:                                  %u\n",
              app->uc_successful_minimization_iterations +
              app->uc_unsuccessful_minimization_iterations +
              app->uc_unknown_minimization_iterations);
      fprintf(stream, "  Successful iterations:                       %u\n",
              app->uc_successful_minimization_iterations);
      fprintf(stream, "  Unsuccessful iterations:                     %u\n",
              app->uc_unsuccessful_minimization_iterations);
      fprintf(stream, "  Unknown iterations:                          %u\n",
              app->uc_unknown_minimization_iterations);
    }
  } else if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
             app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2) {
    fprintf(stream, "Enum a2e:\n");
    fprintf(stdout, "  ENUMA2E A to E irreducible:                  %u\n", app->uc_enuma2e_irreducible);
    fprintf(stdout, "  ENUMA2E A to E non-trivially reducible:      %u\n", app->uc_enuma2e_nontriviallyreducible);
    fprintf(stdout, "  ENUMA2E A to E trivially reducible:          %u\n", app->uc_enuma2e_triviallyreducible);
    fprintf(stdout, "  ENUMA2E A to E unknown reducible:            %u\n", app->uc_enuma2e_unknown);
  }
}


static void
solve_uc_cqqc (QDPLLA2AECCApp * app, QDPLL * qdpll)
{
  qdpll_a2aecc_log(1, app, "begin computing uc\n");
  app->uc_prefix_clausegroupids_cand = app->uc_prefix_clausegroupids;
  app->uc_matrix_clausegroupids_cand = app->uc_matrix_clausegroupids;
  qdpll_a2aecc_log(1, app, "begin deleting irrelevant clause groups\n");
  if (app->uc_atleast_one_clausegroup)
    /* We have to wrap the body below into this if statement to
       prevent an assertion violation in qdpll.c if
       qdpll_get_relevant_clause_groups is called without any clause
       groups having been declared. */
    { 
      if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC &&
          app->options.uc_minimal &&
          app->options.uc_separate_csr)
        delete_irrelevant_clausegroups(app, qdpll, 0, 1);
      else
        delete_irrelevant_clausegroups(app, qdpll, 0, 0);
    }
  else
    qdpll_reset(qdpll);
  qdpll_a2aecc_log(1, app, "end deleting irrelevant clause groups\n");
  if (app->options.uc_minimal) {
    if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC) {
      qdpll_a2aecc_log(1, app, "begin minimizing prefix clause groups\n");
      minimize_clausegroups(app,
                            qdpll,
                            app->uc_prefix_clausegroupids_cand,
                            app->uc_prefix_clausegroupids_next);
      qdpll_a2aecc_log(1, app, "end minimizing prefix clause groups\n");
    }
    if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_C ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC) {
      qdpll_a2aecc_log(1, app, "begin minimizing matrix clause groups\n");
      minimize_clausegroups(app,
                            qdpll,
                            app->uc_matrix_clausegroupids_cand,
                            app->uc_matrix_clausegroupids_next);
      qdpll_a2aecc_log(1, app, "end minimizing matrix clause groups\n");
    }
  }
  qdpll_a2aecc_log(1, app, "end computing uc\n");
  app->core_occurs = (qdpll_a2aecc_core_occurs *)
    qdpll_malloc (qdpll->mm, (app->preamble_max_varid + 1) * sizeof (qdpll_a2aecc_core_occurs));
  memset (app->core_occurs,
          0,
          (app->preamble_max_varid + 1) * sizeof (qdpll_a2aecc_core_occurs));
  if (app->options.print_stats || app->options.uc_filename) {
    qdpll_a2aecc_log(1, app, "begin collecting uc data\n");
    qdpll_a2aecc_collect_data_core (qdpll,
				    app->options.uc_mode,
				    app->options.prefix_as_sets,
				    app->preamble_max_varid,
				    app->in_occurs,
				    app->core_occurs,
				    &app->uc_max_varid,
				    &app->uc_num_clauses,
				    &app->uc_alt_depth);
    qdpll_a2aecc_log(1, app, "end collecting uc data\n");
  }
  if (app->options.print_stats) {
    qdpll_a2aecc_log(1, app, "begin writing statistics\n");
    print_statistics (app, stdout);
    qdpll_a2aecc_log(1, app, "end writing statistics\n");
  }
  if (app->options.uc_filename) {
    FILE *uc_file = fopen (app->options.uc_filename, "w");
    if (!uc_file) {
      print_abort_err (app, "could not open uc file '%s'!\n\n",
                       app->options.uc_filename);
    }
    qdpll_a2aecc_log(1, app, "begin writing uc\n");
    qdpll_a2aecc_print_core (qdpll,
			     uc_file,
			     app->options.uc_mode,
                             app->options.prefix_as_sets,
			     app->preamble_max_varid,
			     app->in_occurs,
			     app->core_occurs,
			     app->uc_max_varid,
			     app->uc_num_clauses);
    qdpll_a2aecc_log(1, app, "end writing uc\n");
    fclose (uc_file);
  }
}


static void
print_enuma2e_witness (QDPLLA2AECCApp * app, QDPLL * qdpll, VarID vid, char c)
{
  qdpll_a2aecc_core_occurs *core_occurs;
  VarID core_max_varid;
  unsigned int core_num_clauses;
  unsigned int core_alt_depth;
  int filename_len;
  char *filename;
  FILE *core_file;
  
  core_occurs = (qdpll_a2aecc_core_occurs *)
    qdpll_malloc (qdpll->mm,
		  (app->preamble_max_varid + 1) * sizeof (qdpll_a2aecc_core_occurs));
  memset (core_occurs,
	  0,
	  (app->preamble_max_varid + 1) * sizeof (qdpll_a2aecc_core_occurs));
  qdpll_a2aecc_log(1, app, "begin collecting enuma2e witness data\n");
  qdpll_a2aecc_collect_data_core (qdpll,
				  app->options.uc_mode,
                                  app->options.prefix_as_sets,
				  app->preamble_max_varid,
				  app->in_occurs,
				  core_occurs,
				  &core_max_varid,
				  &core_num_clauses,
				  &core_alt_depth);
  qdpll_a2aecc_log(1, app, "end collecting enuma2e witness data\n");
  // basename + '.' + VarID + ".1" + ".qdimacs" + \000
  filename_len = strlen(app->options.enuma2e_witness_basename) + 1 + 20 + 2 + 8 + 1;
  filename = qdpll_malloc (qdpll->mm, filename_len * sizeof(char));
  sprintf(filename,
	  "%s.%u.%c.qdimacs",
	  app->options.enuma2e_witness_basename,
	  vid,
	  c);
  core_file = fopen (filename, "w");
  if (!core_file) {
    print_abort_err (app,
		     "could not open enuma2e witness file '%s'!\n\n",
		     filename);
  }
  qdpll_a2aecc_log(1, app, "begin writing enuma2e witness\n");
  qdpll_a2aecc_print_core (qdpll,
			   core_file,
			   app->options.uc_mode,
                           app->options.prefix_as_sets,
			   app->preamble_max_varid,
			   app->in_occurs,
			   core_occurs,
			   core_max_varid,
			   core_num_clauses);
  qdpll_a2aecc_log(1, app, "end writing enuma2e witness\n");
  fclose (core_file);
  qdpll_free(qdpll->mm, filename, filename_len * sizeof(char));
  qdpll_free(qdpll->mm,
	     core_occurs,
	     (app->preamble_max_varid + 1) * sizeof(qdpll_a2aecc_core_occurs));
}


static void
deactivate_nonoccurring_clausegroups (QDPLLA2AECCApp * app,
                                      QDPLL *qdpll,
                                      ClauseGroupID *beg,
                                      ClauseGroupID *end,
                                      cgidhash_t *cgidh1,
                                      cgidhash_t *cgidh2)
{
  unsigned int num_clausegroups_deactivated = 0;

  for (ClauseGroupID *cgid = beg;
       cgid < end;
       cgid++) {
    if (*cgid == 0)
      continue;
    if (!cgidhash_present(cgidh1, *cgid) &&
        (cgidh2 == NULL || !cgidhash_present(cgidh2, *cgid))) {
      if (!qdpll_exists_clause_group(qdpll, *cgid)) {
        assert(0);
        QDPLL_A2AECC_ABORT_APP (1, "internal error: clause group does not exist!\n");
        return;
      }
      if (qdpll_isactive_clause_group(qdpll, *cgid)) {
        num_clausegroups_deactivated++;
        qdpll_deactivate_clause_group(qdpll, *cgid);
      }
    }
  }
  qdpll_a2aecc_log(1, app, "deactivated %d clause groups\n", num_clausegroups_deactivated);
}


static void
solve_uc_enuma2e (QDPLLA2AECCApp * app, QDPLL * qdpll)
{
  qdpll_a2aecc_log(1, app, "begin computing enuma2e1/2\n");
  qdpll_reset(qdpll);
  for (app->uc_prefix_clausegroupids_cand = app->uc_prefix_clausegroupids,
         app->uc_varid_to_a2aeccvarid_cand = app->uc_varid_to_a2aeccvarid;
       app->uc_prefix_clausegroupids_cand < app->uc_prefix_clausegroupids_next;
       app->uc_prefix_clausegroupids_cand++,
         app->uc_varid_to_a2aeccvarid_cand += 2) {
    QDPLLResult tmpres;
    ClauseGroupID cgid;
    VarID vid;
        
    if (app->uc_varid_to_a2aeccvarid_cand >= app->uc_varid_to_a2aeccvarid_next) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: end of uc_varid_to_a2aeccvarid reached!\n");
      return;
    }
    vid = *app->uc_varid_to_a2aeccvarid_cand;

    cgid = *app->uc_prefix_clausegroupids_cand;
    /* Dummy element, marks an A quantified variable without
       occurrences in the matrix. */
    if (cgid == 0) {
      app->uc_enuma2e_triviallyreducible++;
      fprintf(stdout, "varid %d trivially A to E reducible\n", vid);
      continue;
    }
    if (!qdpll_exists_clause_group(qdpll, cgid)) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: clause group does not exist!\n");
      return;
    }
    if (!qdpll_isactive_clause_group(qdpll, cgid)) {
      assert(0);
      QDPLL_A2AECC_ABORT_APP (1, "internal error: clause group is not active!\n");
      return;
    }

    qdpll_deactivate_clause_group(qdpll, cgid);
    qdpll_a2aecc_log(1, app, "begin checking satisfiability of weakened formula\n");
    tmpres = qdpll_sat(qdpll);
    qdpll_a2aecc_log(1, app, "end checking satisfiability of weakened formula\n");
    qdpll_reset(qdpll);
    if (app->options.enuma2e_witness_basename) {
      // If vid is not A to E reducible, then this is SAT.
      // If vid is (non-)trivially A to E reducible, then this is UNSAT.
      qdpll_a2aecc_log(1, app, "begin printing enuma2e witness\n");
      print_enuma2e_witness (app, qdpll, vid, '1');
      qdpll_a2aecc_log(1, app, "end printing enuma2e witness\n");
    }
    if (tmpres == QDPLL_RESULT_UNSAT) {
      if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2 &&
          app->in_occurs[vid].matrix_pos &&
          app->in_occurs[vid].matrix_neg) {
        /* Minimize prefix and minimize clauses in which vid doesn't occur. */
        cgidhash_t *cgidh;
        ClauseGroupID *cgid2;
        if (app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids != NULL &&
            (app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids_next !=
             app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids)) {
          cgidh = cgidhash_new(qdpll->mm,
                               2 *
                               (app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids_next -
                                app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids));
          for (cgid2 = app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids;
               cgid2 < app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids_next;
               cgid2++) {
            if (cgid2 >= app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids_end) {
              assert(0);
              QDPLL_A2AECC_ABORT_APP (1, "internal error: end of uc_varid_to_varocc_clausegroupids reached!\n");
              return;
            }
            if (!qdpll_exists_clause_group(qdpll, *cgid2)) {
              assert(0);
              QDPLL_A2AECC_ABORT_APP (1, "internal error: clause group does not exist!\n");
              return;
            }
            if (*cgid2 == 0) {
              assert(0);
              QDPLL_A2AECC_ABORT_APP (1, "internal error: clause group id is 0!\n");
              return;
            }
            cgidhash_insert(cgidh, *cgid2);
          }
        } else {
          cgidh = cgidhash_new(qdpll->mm, 0);
        }
        for (cgid2 = app->uc_prefix_clausegroupids;
             cgid2 < app->uc_prefix_clausegroupids_next;
             cgid2++) {
          if (cgid2 >= app->uc_prefix_clausegroupids_end) {
            assert(0);
            QDPLL_A2AECC_ABORT_APP (1, "internal error: end of uc_prefix_clausegroupids reached!\n");
            return;
          }
          if (*cgid2 == 0 || *cgid2 == cgid)
            continue;
          if (!qdpll_exists_clause_group(qdpll, *cgid2)) {
            assert(0);
            QDPLL_A2AECC_ABORT_APP (1, "internal error: clause group does not exist!\n");
            return;
          }
          if (qdpll_isactive_clause_group(qdpll, *cgid2)) {
            qdpll_deactivate_clause_group(qdpll, *cgid2);
            qdpll_a2aecc_log(1, app, "begin checking satisfiability of weakened formula\n");
            tmpres = qdpll_sat(qdpll);
            qdpll_a2aecc_log(1, app, "end checking satisfiability of weakened formula\n");
            if (tmpres == QDPLL_RESULT_UNSAT) {
              /* Clause set refinement. */
              ClauseGroupID *relevant_cgids;
              cgidhash_t *cgidh2;
              relevant_cgids = qdpll_get_relevant_clause_groups(qdpll);
              qdpll_reset(qdpll);
              cgidh2 = fill_cgidh_with_relevant_clausegroups(relevant_cgids);
              deactivate_nonoccurring_clausegroups(app,
                                                   qdpll, 
                                                   cgid2 + 1,
                                                   app->uc_prefix_clausegroupids_next,
                                                   cgidh2,
                                                   NULL);
              deactivate_nonoccurring_clausegroups(app,
                                                   qdpll, 
                                                   app->uc_matrix_clausegroupids,
                                                   app->uc_matrix_clausegroupids_next,
                                                   cgidh,
                                                   cgidh2);
              cgidhash_delete(qdpll->mm, cgidh2);
              free(relevant_cgids);
            } else {
              qdpll_reset(qdpll);
              qdpll_activate_clause_group(qdpll, *cgid2);
            }
          }
        }
        for (cgid2 = app->uc_matrix_clausegroupids;
             cgid2 < app->uc_matrix_clausegroupids_next;
             cgid2++) {
          if (cgid2 >= app->uc_matrix_clausegroupids_end) {
            assert(0);
            QDPLL_A2AECC_ABORT_APP (1, "internal error: end of uc_matrix_clausegroupids reached!\n");
            return;
          }
          if (!qdpll_exists_clause_group(qdpll, *cgid2)) {
            assert(0);
            QDPLL_A2AECC_ABORT_APP (1, "internal error: clause group does not exist!\n");
            return;
          }
          if ((!cgidhash_present(cgidh, *cgid2)) && qdpll_isactive_clause_group(qdpll, *cgid2)) {
            qdpll_deactivate_clause_group(qdpll, *cgid2);
            qdpll_a2aecc_log(1, app, "begin checking satisfiability of weakened formula\n");
            tmpres = qdpll_sat(qdpll);
            qdpll_a2aecc_log(1, app, "end checking satisfiability of weakened formula\n");
            if (tmpres == QDPLL_RESULT_UNSAT) {
              /* Clause set refinement. */
              ClauseGroupID *relevant_cgids;
              cgidhash_t *cgidh2;
              relevant_cgids = qdpll_get_relevant_clause_groups(qdpll);
              qdpll_reset(qdpll);
              cgidh2 = fill_cgidh_with_relevant_clausegroups(relevant_cgids);
              deactivate_nonoccurring_clausegroups(app,
                                                   qdpll, 
                                                   cgid2 + 1,
                                                   app->uc_matrix_clausegroupids_next,
                                                   cgidh,
                                                   cgidh2);
              cgidhash_delete(qdpll->mm, cgidh2);
              free(relevant_cgids);
            } else {
              qdpll_reset(qdpll);
              qdpll_activate_clause_group(qdpll, *cgid2);
            }
          }
        }
        cgidhash_delete(qdpll->mm, cgidh);
	if (app->options.enuma2e_witness_basename) {
	  // This is UNSAT.
	  qdpll_a2aecc_log(1, app, "begin printing enuma2e witness\n");
	  print_enuma2e_witness (app, qdpll, vid, '2');
	  qdpll_a2aecc_log(1, app, "end printing enuma2e witness\n");
	}
      }
      if (app->in_occurs[vid].matrix_pos &&
          app->in_occurs[vid].matrix_neg &&
          app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids != NULL &&
          (app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids_next !=
           app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids)) {
        ClauseGroupID *cgid2;
        for (cgid2 = app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids;
             cgid2 < app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids_next;
             cgid2++) {
          if (cgid2 >= app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids_end) {
            assert(0);
            QDPLL_A2AECC_ABORT_APP (1, "internal error: end of uc_varid_to_varocc_clausegroupids reached!\n");
            return;
          }
          if (!qdpll_exists_clause_group(qdpll, *cgid2)) {
            assert(0);
            QDPLL_A2AECC_ABORT_APP (1, "internal error: clause group does not exist!\n");
            return;
          }
          if (!qdpll_isactive_clause_group(qdpll, *cgid2)) {
            assert(0);
            QDPLL_A2AECC_ABORT_APP (1, "internal error: clause group is not active!\n");
            return;
          }
          qdpll_deactivate_clause_group(qdpll, *cgid2);
        }
        qdpll_a2aecc_log(1, app, "begin checking satisfiability of weakened formula\n");
        tmpres = qdpll_sat(qdpll);
        qdpll_a2aecc_log(1, app, "end checking satisfiability of weakened formula\n");
        qdpll_reset(qdpll);
	if (app->options.enuma2e_witness_basename) {
	  // If vid is non-trivially A to E reducible, then this is SAT.
	  // If vid is trivially A to E reducible, then this is UNSAT.
	  qdpll_a2aecc_log(1, app, "begin printing enuma2e witness\n");
	  print_enuma2e_witness (app, qdpll, vid, '3');
	  qdpll_a2aecc_log(1, app, "end printing enuma2e witness\n");
	}
        if (tmpres == QDPLL_RESULT_SAT) {
          app->uc_enuma2e_nontriviallyreducible++;
          fprintf(stdout, "varid %d non-trivially A to E reducible\n", vid);
        } else if (tmpres == QDPLL_RESULT_UNSAT) {
          // Here we really can only say that vid is either trivially
          // or non-trivially A to E reducible - we just may not have
          // found the proper witness for non-trivial A to E
          // reducibility.
          app->uc_enuma2e_triviallyreducible++;
          fprintf(stdout, "varid %d trivially A to E reducible\n", vid);
        } else {
          app->uc_enuma2e_unknown++;
          fprintf(stdout, "varid %d unknown A to E reducible\n", vid);
        }
        qdpll_activate_clause_group(qdpll, cgid);
        for (cgid2 = app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids;
             cgid2 < app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids_next;
             cgid2++) {
          qdpll_activate_clause_group(qdpll, *cgid2);
        }
      } else {
        app->uc_enuma2e_triviallyreducible++;
        fprintf(stdout, "varid %d trivially A to E reducible\n", vid);
        qdpll_activate_clause_group(qdpll, cgid);
      }
      if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2 &&
          app->in_occurs[vid].matrix_pos &&
          app->in_occurs[vid].matrix_neg) {
        ClauseGroupID *cgid2;
        for (cgid2 = app->uc_prefix_clausegroupids;
             cgid2 < app->uc_prefix_clausegroupids_next;
             cgid2++) {
          if (*cgid2 == 0)
            continue;
          if (!qdpll_isactive_clause_group(qdpll, *cgid2))
            qdpll_activate_clause_group(qdpll, *cgid2);
        }
        for (cgid2 = app->uc_matrix_clausegroupids;
             cgid2 < app->uc_matrix_clausegroupids_next;
             cgid2++) {
          if (!qdpll_isactive_clause_group(qdpll, *cgid2))
            qdpll_activate_clause_group(qdpll, *cgid2);
        }
      }
    } else {
      if (tmpres == QDPLL_RESULT_SAT) {
        app->uc_enuma2e_irreducible++;
        fprintf(stdout, "varid %d not A to E reducible\n", vid);
      } else {
        app->uc_enuma2e_unknown++;
        fprintf(stdout, "varid %d unknown A to E reducible\n", vid);
      }
      qdpll_activate_clause_group(qdpll, cgid);
    }
  }
  qdpll_a2aecc_log(1, app, "end computing enuma2e1/2\n");
  if (app->options.print_stats) {
    qdpll_a2aecc_log(1, app, "begin writing statistics\n");
    print_statistics (app, stdout);
    qdpll_a2aecc_log(1, app, "end writing statistics\n");
  }
}


static QDPLLResult
solve (QDPLLA2AECCApp * app, QDPLL * qdpll)
{
  QDPLLResult result = QDPLL_RESULT_UNKNOWN;
  
  qdpll_a2aecc_log(1, app, "begin checking satisfiability of formula\n");
  result = qdpll_sat(qdpll);
  qdpll_a2aecc_log(1, app, "end checking satisfiability of formula\n");

  if (result == QDPLL_RESULT_UNSAT) {
    if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_C ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
        app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC) {
      solve_uc_cqqc(app, qdpll);
    } else if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
               app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2) {
      solve_uc_enuma2e(app, qdpll);
    } else {
      qdpll_reset(qdpll);
    }
  } else {
    qdpll_reset(qdpll);
  }

  if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_NONE &&
      app->options.print_stats) {
    qdpll_a2aecc_log(1, app, "begin writing statistics\n");
    print_statistics (app, stdout);
    qdpll_a2aecc_log(1, app, "end writing statistics\n");
  }

  return result;
}


static void
print_result_message (QDPLLA2AECCApp * app,
                      QDPLL * qdpll,
                      QDPLLResult result,
                      FILE *stream)
{
  /* Print result message; this may not always be desired and depends on the 
     current usage of the solver library. */
  if (!app->options.qdimacs_output)
    {
      /* Print own output format. */
      if (result == QDPLL_RESULT_SAT)
        fprintf (stream, "SAT\n");
      else if (result == QDPLL_RESULT_UNSAT)
        fprintf (stream, "UNSAT\n");
      else if (!app->options.pretty_print && !app->options.deps_only)
        {
          assert (result == QDPLL_RESULT_UNKNOWN);
          fprintf (stream, "UNKNOWN\n");
        }
    }
  else
    {
      /* Print qdimacs-compliant output format. */
      qdpll_print_qdimacs_output (qdpll);
    }
}


static void
print_usage ()
{
  fprintf (stdout, USAGE1);
  fprintf (stdout, USAGE2);
  fprintf (stdout, USAGE3);
}


static void
print_version ()
{
  fprintf (stderr, VERSION);
}


static void
cleanup (QDPLL * qdpll, QDPLLA2AECCApp * app)
{
  if (app->options.in_filename)
    fclose (app->options.in);
  if (app->options.a2aecc_filename)
    fclose (app->options.a2aecc_file);

  if (app->comments)
    qdpll_free(qdpll->mm,
               app->comments,
               (app->max_len_comments) * sizeof(char));
  if (app->in_occurs)
    qdpll_free(qdpll->mm,
               app->in_occurs,
               (app->preamble_max_varid + 1) * sizeof(qdpll_a2aecc_in_occurs));
  if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2 ||
      app->options.a2aecc_filename)
    if (app->uc_prefix_clausegroupids)
      qdpll_free(qdpll->mm,
                 app->uc_prefix_clausegroupids,
                 app->preamble_max_varid * sizeof (ClauseGroupID));
  if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_C ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2)
    if (app->uc_matrix_clausegroupids)
      qdpll_free(qdpll->mm,
                 app->uc_matrix_clausegroupids,
                 app->preamble_num_clauses * sizeof (ClauseGroupID));
  if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_Q ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_QC ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1 ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2 ||
      app->options.a2aecc_filename)
    if (app->uc_varid_to_a2aeccvarid)
      qdpll_free(qdpll->mm,
                 app->uc_varid_to_a2aeccvarid,
                 2 * app->preamble_max_varid * sizeof (VarID));
  if (app->core_occurs)
    qdpll_free(qdpll->mm,
               app->core_occurs,
               (app->preamble_max_varid + 1) * sizeof(qdpll_a2aecc_core_occurs));
  if (app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E1  ||
      app->options.uc_mode == QDPLL_A2AECC_UC_MODE_ENUMA2E2) {
    if (app->uc_varid_to_varocc_clausegroupids) {
      for (VarID vid = 0; vid <= app->preamble_max_varid; vid++)
        if (app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids)
          qdpll_free(qdpll->mm,
                     app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids,
                     (app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids_end -
                      app->uc_varid_to_varocc_clausegroupids[vid].varocc_clausegroupids) *
                     sizeof (ClauseGroupID));
      qdpll_free (qdpll->mm,
                  app->uc_varid_to_varocc_clausegroupids,
                  (app->preamble_max_varid + 1) * sizeof (QDPLLA2AECCVarOccCClauseGroupIDs));
    }
  }

  qdpll_delete (qdpll);
}


int
qdpll_a2aecc_main (int argc, char **argv)
{
  QDPLLResult result = QDPLL_RESULT_UNKNOWN;
  QDPLLA2AECCApp app;
  qdplla2aeccapp_init (&app);

  qdpll = qdpll_create ();

  parse_cmd_line_options (&app, qdpll, argc, argv);
  check_options (&app);
  set_signal_handlers (&app);

  if (app.options.print_usage)
    {
      print_usage ();
      cleanup (qdpll, &app);
      return result;
    }
  if (app.options.print_version)
    {
      print_version ();
      cleanup (qdpll, &app);
      return result;
    }

  if (app.options.max_time)
    alarm (app.options.max_time);

  qdpll_a2aecc_log(1, &app, "begin parsing input formula\n");
  parse (&app, qdpll, app.options.in, app.options.trace);
  qdpll_a2aecc_log(1, &app, "end parsing input formula\n");

  if (app.options.pretty_print)
    {
      /* Call 'qdpll_gc' to clean up the formula by removing variables
         which have no occurrences and removing empty quantifier
         blocks. */
      qdpll_gc (qdpll);
      qdpll_a2aecc_log(1, &app, "begin pretty printing input formula\n");
      qdpll_print (qdpll, stdout);
      qdpll_a2aecc_log(1, &app, "end pretty printing input formula\n");
      if (app.options.print_stats) {
        qdpll_a2aecc_log(1, &app, "begin writing statistics\n");
        print_statistics (&app, stdout);
        qdpll_a2aecc_log(1, &app, "end writing statistics\n");
      }
    }
  else if (app.options.a2aecc_filename)
    {
      qdpll_a2aecc_log(1, &app, "begin writing a2aecc formula\n");
      fprintf(app.options.a2aecc_file, "c BEGIN changes made by a2aecc\n");
      for (VarID *vidp = app.uc_varid_to_a2aeccvarid,
             *vide = app.uc_varid_to_a2aeccvarid_next;
           vidp < vide;
           vidp += 2) {
        VarID vid = *vidp;
        if (vid == 0) {
          assert(0);
          QDPLL_A2AECC_ABORT_APP (1, "internal error: vid is 0!\n");
          return result;
        }
        VarID vidaux = *(vidp+1);
        if (vidaux == 0) {
          assert(0);
          QDPLL_A2AECC_ABORT_APP (1, "internal error: vidaux is 0!\n");
          return result;
        }
        if (app.in_occurs[vid].matrix_pos || app.in_occurs[vid].matrix_neg) {
          if ((app.in_occurs[vid].matrix_pos &&
               app.in_occurs[vid].matrix_neg) ||
              !app.options.poldep_clauses) {
            fprintf(app.options.a2aecc_file,
                    "c changed \"a %d\" to \"a %d e %d (%d, %d) (%d, %d)\"\n",
                    vid,
                    vidaux,
                    vid,
                    -vidaux,
                    vid,
                    vidaux,
                    -vid);
          } else if (app.in_occurs[vid].matrix_neg) {
            fprintf(app.options.a2aecc_file,
                    "c changed \"a %d\" to \"a %d e %d (%d, %d)\"\n",
                    vid,
                    vidaux,
                    vid,
                    -vidaux,
                    vid);
          } else if (app.in_occurs[vid].matrix_pos) {
            fprintf(app.options.a2aecc_file,
                    "c changed \"a %d\" to \"a %d e %d (%d, %d)\"\n",
                    vid,
                    vidaux,
                    vid,
                    vidaux,
                    -vid);
          }
        } else {
          fprintf(app.options.a2aecc_file,
                  "c changed \"a %d\" to \"a %d e %d\"\n",
                  vid,
                  vidaux,
                  vid);
        }
      }
      fprintf(app.options.a2aecc_file, "c END changes made by a2aecc\n");
      fprintf(app.options.a2aecc_file, "%s", app.comments);
      qdpll_print (qdpll, app.options.a2aecc_file);
      qdpll_a2aecc_log(1, &app, "end writing a2aecc formula\n");
      if (app.options.print_stats) {
        qdpll_a2aecc_log(1, &app, "begin writing statistics\n");
        print_statistics (&app, stdout);
        qdpll_a2aecc_log(1, &app, "end writing statistics\n");
      }
      cleanup (qdpll, &app);
      return result;
    }
  else if (app.options.deps_only)
    {
      qdpll_init_deps (qdpll);
      if (app.options.print_deps)
        print_deps (&app, qdpll);
      if (app.options.dump_dep_graph)
        qdpll_dump_dep_graph (qdpll);
      if (app.options.print_stats) {
        qdpll_a2aecc_log(1, &app, "begin writing statistics\n");
        print_statistics (&app, stdout);
        qdpll_a2aecc_log(1, &app, "end writing statistics\n");
      }
    }
  else
    {
      qdpll_a2aecc_log(1, &app, "begin solving formula\n");
      result = solve (&app, qdpll);
      qdpll_a2aecc_log(1, &app, "end solving formula\n");
#if (COMPUTE_STATS || COMPUTE_TIMES)
        qdpll_print_stats (qdpll);
#endif
    }

  if (app.options.trace == TRACE_QRP)
    fprintf (stdout, "r ");
  else if (app.options.trace == TRACE_BQRP)
    fprintf (stdout, "%cr ", 0);

  print_result_message (&app, qdpll, result, stdout);

  cleanup (qdpll, &app);

  return result;
}


int
main (int argc, char **argv)
{
  return qdpll_a2aecc_main (argc, argv);
}
